#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "aap_aitu.h"
#include "allocation.h"
#include "defender.h"

/*------------------------------------------------------------------*/
struct aitu_raw {
    unsigned int current_time;
    addr_range_raw ranges[500];
};
/*------------------------------------------------------------------*/
static int send_msg(aap_aitu *a) {

    aap_msg *msg;
    struct aitu_raw aitu;
    int len, i;
    list *allocs;
  
    /* construct the message body */
    bzero(&aitu, sizeof(struct aitu_raw));

    /* insert current time */
    aitu.current_time = htonl(time(NULL));

    allocs = a->allocations;

    /* copy address ranges into message body */
    for(i = 0; i < allocs->length(allocs); i++) {
        addr_range *ar;
        allocation *alloc = allocs->elem_at(allocs, i);
        ar = alloc->range(alloc);
        aitu.ranges[i] = ar->raw(ar);
    }
    /* total length of the message body */
    len = i * sizeof(addr_range_raw) + 4;
  
    /* only send message if there are allocations */
    if(i > 0) {
        msg = aap_msg_new();
        msg->create(msg, AAP_MSG_AITU, 0 /* XXX */,
                    0, 0, (char *)&aitu, len);
        msg->send(msg, a->n, a->ns);
    }
    return 0;
}

/*------------------------------------------------------------------*/

int aap_aitu_receive(aap *aap, aap_msg *msg) {
    int len, i, j, k, pos, defending;
    allocation *alloc;
    list *l;
    defender *def;
  
    struct aitu_raw *aitu;
    unsigned int clock_skew;
    int rseq, mseq, addrfamily;
  
    rseq = msg->rseq(msg);
    mseq = msg->mseq(msg);
    addrfamily = msg->addrfamily(msg);
    
    msg->body(msg, (char **)&aitu, &len);

    /* adjust for clock skew as in section
       5.3.1 of draft-ietf-malloc-aap-03.txt */
    clock_skew = ntohl(aitu->current_time) - time(NULL);
  
    /* i should contain the number of ranges */
    i = (len - 4) / sizeof(addr_range_raw);
    if(len != (i * sizeof(addr_range_raw) + 4)) {
        /* malformed message */
        DEBUG('d', "aap_aitu: Received malformed message.\n");
        return 1;
    }

    /* process the address ranges */
    /* check if the addr_range overlaps an existing allocation */
    /* check if there are outstanding ACLMs to discard if overlaps*/
    /* update Allocation record */
    /* check unanswered_aclm to see if the address is claimed */
    
    while(i > 0) {
        addr_range *ar;
        list *overlaps;

        ar = addr_range_new_from_raw(&(aitu->ranges[i - 1]));
        ar->debug_print(ar);
        ar->adjust(ar, clock_skew);

        /* check to see if the incomming ACLM is a reclaim for an
           address range that we are defending. if so, the
           defens should be canceled */
        l = list_new();
        len = aap->defens_list->length(aap->defens_list);
        for(j = 0; j < len; j++) {
            defender *def = aap->defens_list->elem_at(aap->defens_list, j);
            if(def->match(def, rseq, msg->sender(msg))) {
                l->insert(l, def);
            }
        }

        len = l->length(l);
        for(j = 0; j < len; j++) {
            defender *def = l->elem_at(l, j);
            def->cancel(def);
            pos = aap->defens_list->find(aap->defens_list, def);
            aap->defens_list->remove(aap->defens_list, pos);
            def->delete(&def);
            def = NULL;
        }

        overlaps = aap->arec->overlap(aap->arec, ar);

        /* check if the incomming ACLM overlaps an allocation
           in the allocation record, in which case it should be
           defended. */
        defending = 0;
        for(k = 0; k < aap->defens_list->length(aap->defens_list); k++) {
            def = aap->defens_list->elem_at(aap->defens_list, k);
            if(def->overlap(def, ar)) {
                defending++;
            }
      
        }
        if(!defending) {
            if(!overlaps->is_empty(overlaps)) {      
                for( j = 0 ; j < overlaps->length(overlaps) ; j++) {
                    allocation *tmp_alloc = overlaps->elem_at(overlaps, j);
                    
                    if((tmp_alloc->owner_addr(tmp_alloc) != msg->sender(msg)) &&
                       (tmp_alloc->flag != ALLOCATION_PREALLOC)) {
                        /* Above added by Fredrik. Don't defend preallocations. */
                        def = defender_new(aap, tmp_alloc, rseq, msg->sender(msg));	  
                        def->defend(def);
                        def->aap->defens_list->insert(def->aap->defens_list, def);
                    }
                }
            }
            overlaps->delete(overlaps);
            overlaps = NULL;
        }
        
        
        if((overlaps != NULL) && (!overlaps->is_empty(overlaps))){
            for(j = 0; j < overlaps->length(overlaps) ; j++){
                allocation *tmp_alloc = overlaps->elem_at(overlaps, j);
                if(tmp_alloc->flag == ALLOCATION_PREALLOC){
                    /* AITU always overrides a previous AITU. */
                    /* Delete earlier allocation and insert the new one. */
                    
                    aap->arec->remove(aap->arec, tmp_alloc);
                    tmp_alloc->delete(&tmp_alloc);
                    
                    alloc = allocation_new(ALLOCATION_PREALLOC,
                                           ar,
                                           time(NULL),
                                           msg->sender(msg));
                    aap->arec->insert(aap->arec, alloc);
                }
            }
        } else {
            /* a new allocation to be inserted into the allocation record */
            aap->arec->insert(aap->arec, allocation_new(ALLOCATION_PREALLOC,
                                                        ar,
                                                        time(NULL),
                                                        msg->sender(msg)));
        }
        
        if(overlaps != NULL) {
            overlaps->delete(overlaps);
            overlaps = NULL;
        }

        i--;
    }
  
    return 0;

}
/*------------------------------------------------------------------*/
static void delete(aap_aitu *a) {
    free(a);
}
/*------------------------------------------------------------------*/
aap_aitu *aap_aitu_new(net *n, net_socket *ns, tim *t, cfg *c, aap *aap, list *allocs, int rseq) {
  
    aap_aitu *aitu = malloc(sizeof(aap_aitu));

    aitu->n = n;
    aitu->ns = ns;
    aitu->t = t;
    aitu->c = c;
    aitu->aap = aap;
    aitu->allocations = allocs;
    aitu->rseq = rseq;
    aitu->mseq = 0;
  
    /*aitu->send = send_msg;*/
    aitu->delete = delete;
  

    return aitu;
}
/*------------------------------------------------------------------*/
