#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include "aap_asa.h"
#include "aap.h"
#include "debug.h"

struct asa_raw {
    unsigned int current_time;
    unsigned int exp_time;
    addr_range_raw ranges[500];
};

static int send_asa(aap_asa *a, list *ranges);
static void resend_asa(void *obj);
static void asa_resend_timer(void *obj);

/*-----------------------------------------------------------*/

/**
 *  If the MAAS receives an ASA message whose Expiration Time (after
 *  adjusting for clock skew, as described in section 5.3.1) is newer
 *  than the last ASA message processed and the newer ASA message does
 *  not indicate that all of the addresses being claimed are available,
 *  the MAAS MUST cancel the Claim Timer and give up on the unavailable
 *  addresses. This should not normally happen, since the MAAS should
 *  have been aware of the set of available addresses before it sent out
 *  the ACLM message and the set of available addresses should not change
 *  in a way that invalidates outstanding claims.
*/

int aap_asa_receive(aap *aap, aap_msg *msg){

    unsigned int clock_skew;
    struct asa_raw *asa;
    int len, i;
    addr_range *ar = (addr_range *)malloc(sizeof(addr_range));
    unsigned int expiration_time, last_exp_time;

    msg->body(msg, (char **)&asa, &len);

    /* i should contain the number of ranges */
    i = (len - 8) / sizeof(addr_range_raw);
    if(len != (i * sizeof(addr_range_raw) + 8)) {
        /* malformed message */
        DEBUG('d', "aap_asa: Received malformed message.");
        return 1;
    }
 
    aap->last_asa_receive_time = time(NULL);

    /* Am I a Prefix Coordinator or a MAAS? */
    if(aap->net->is_self(aap->net, msg->sender(msg))){
        DEBUG('d',"Ignoring ASA sent by self.\n");
        /* Restart resend timer. */
        aap->asa_resend_timer = 0;
        asa_resend_timer(aap);
    } else if( aap->cfg->type == CFG_PC ){
        /* since I'm a PC, do nothing...for now */
        DEBUG('d',"PC received ASA...\n");
    }else{
        /* check Expiration time of the incoming ASA */
        clock_skew = ntohl(asa->current_time) - time(NULL);
        expiration_time = ntohl(asa->exp_time) + clock_skew;
        last_exp_time =  aap->last_asa_exp_time;

       /* insert the asa into asa_ranges if it is fresher than the last msg */
        
        if( expiration_time > last_exp_time ){
            char *tmp;
            aap->last_asa_exp_time = expiration_time;
            msg->body(msg, &tmp, &aap->last_asa_body_len);
            memcpy(aap->last_asa_body, tmp, aap->last_asa_body_len);
            
            /* Reset ASA expiration timer and discard ASA resend timer. */
            DEBUG('d', "Reset ASA expiration timer and discard ASA resend timer.\n");
            if(aap->asa_exp_timer != 0){
                aap->tim->unset(aap->tim, aap->asa_exp_timer);
            }
            aap->asa_exp_timer = aap->tim->set(aap->tim,
                                               expiration_time - time(NULL),
                                               (tim_callback) asa_resend_timer,
                                               aap);
            if(aap->asa_resend_timer != 0){
                aap->tim->unset(aap->tim, aap->asa_resend_timer);
            }
          
            while(i > 0) {
                int j;
                plist *asa_ranges;
	
                ar = addr_range_new_from_raw(&(asa->ranges[i - 1]));
                ar->adjust(ar, clock_skew);
	
                /* before inserting the ASA, check if there exists
                   an ASA for that addr_range.
                */
                asa_ranges = aap->arec->asa_ranges;
                for(j = 0 ; j < asa_ranges->length(asa_ranges) ; j++){

                    addr_range *tmp_ar;

                    tmp_ar = asa_ranges->elem_at(asa_ranges, j);

                    /* if there already exists a range
                       in arec, substitute it with the new */
                    if( tmp_ar->overlap(tmp_ar, ar) ){
	    
                        /* first remove the old range */
                        aap->arec->asa_ranges->remove(aap->arec->asa_ranges, j);
                        aap->arec->asa_end->remove(aap->arec->asa_end,
                                                   tmp_ar->end_time(tmp_ar));
                        tmp_ar->delete(tmp_ar);

                        /* then insert the new one */
                        aap->arec->insert_asa(aap->arec, ar);
                        break;
                    }	  
                }
	
                i--;
            }

            /* set Expiration timer in arec */
            
        } else {
            /* Reset the ASA resend timer since another MAAS has resent. */
            asa_resend_timer(aap);
        }         

    }  
  
    return 0;
}


/*-----------------------------------------------------------*/

static int send_asa(aap_asa *a, list *ranges){

    aap_msg *msg;
    struct asa_raw asa;
    int len, i;
    /*list *allocs;*/
  
    /* construct the message body */
    bzero(&asa, sizeof(struct asa_raw));

    /* insert current time */
    asa.current_time = htonl(time(NULL));

    /*
      All ASA messages include an Expiration Time. Normally, the Expiration
      Time is set so that several newer ASA messages will have been sent
      before an ASA message expires.
    */
    asa.exp_time = htonl(time(NULL) + 10000);

  
    /* copy address ranges into message body */
    for(i = 0; i < ranges->length(ranges); i++) {
        addr_range *ar;
        ar = ranges->elem_at(ranges, i);
        asa.ranges[i] = ar->raw(ar);
    }
    /* total length of the message body */
    len = i * sizeof(addr_range_raw) + 8; /* current_time + expiration time */
  
    /* only send message if there are allocations */
    if(i > 0) {
        msg = aap_msg_new();
        msg->create(msg, AAP_MSG_ASA, AAP_MSG_IP,
                    a->rseq, a->mseq, (char *)&asa, len);
        msg->send(msg, a->n, a->ns);
        a->mseq++;
    }

    return 0;
  
  
}

/*-----------------------------------------------------------*/

/*
 * Callback function used to resend last received ASA message when the ASA
 * resend timer has fired.
 */
static void resend_asa(void *obj){
    unsigned int current_time, old_curr_time;
    aap_msg *msg;
    aap *aap = obj;
    
    DEBUG('d',"ASA resend timer has fired, resending ASA.\n");
    /* "If the timer expires without an ASA message arriving, the MAAS SHOULD
       send an ASA message with the same contents as the expired ASA message,
       but with the Current Time field incremented by the amount of time passed
       since the expired ASA message was received." */
    memcpy(&old_curr_time, aap->last_asa_body, 4);
    old_curr_time = ntohl(old_curr_time);
    current_time = old_curr_time + (time(NULL) - aap->last_asa_receive_time);
    current_time = htonl(current_time);

    /* Copy new current time to old message body. */
    memcpy(aap->last_asa_body, &current_time, 4);

    /* Create and send ASA message. */
    msg = aap_msg_new();
    msg->create(msg, AAP_MSG_ASA, AAP_MSG_IP,
                aap->next_rseq(aap), 0, aap->last_asa_body,
                aap->last_asa_body_len);
    msg->send(msg, aap->net, aap->out_ns);
}

/*
 * Callback function used to start the ASA resend timer when the expiration
 * timer of the last received ASA message has fired.
 *
 * Also used to reset the resend timer when another MAAS has resent an ASA.
 */
static void asa_resend_timer(void *obj){
    int interval;
    double random;
    aap *aap = obj;


    if ((aap->asa_exp_timer == 0) && (aap->asa_resend_timer == 0)){
        /* This MAAS has resent an ASA, restart the resend timer. */
        DEBUG('d',"Restarting ASA resend timer.\n");
    } else if (aap->asa_exp_timer != 0){
        /* ASA expiration timer has fired, reset timer id. */
        DEBUG('d',"ASA expiration timer has fired.\n");
        aap->asa_exp_timer = 0;
    } else if (aap->asa_resend_timer != 0){
        /* Another MAAS has resent an ASA, remove resend timer. */
        DEBUG('d',"Another MAAS has resent an ASA, restarting resend timer.\n");
        aap->tim->unset(aap->tim, aap->asa_resend_timer);
    }
    
    /* Create an interval between [ASA-INTERVAL]*0.7 and [ASA-INTERVAL]*1.3. */
    srand(time(NULL));
    random = ((double)rand()) / ( 1.66667 * (double)RAND_MAX) + 0.7;
    interval = (int) ((aap->cfg->ASA_INTERVAL) * random);

    /* Set ASA resend timer. */
    aap->asa_resend_timer = aap->tim->set(aap->tim, interval,
                                          (tim_callback) resend_asa, aap);
}

/*------------------------------------------------------------------*/
static void delete(aap_asa *a) {
    free(a);
}
/*-----------------------------------------------------------*/

aap_asa *aap_asa_new(net *n, net_socket *ns, tim *t, cfg *c, aap *aap, int rseq){

    aap_asa *asa = malloc(sizeof(aap_asa));

    asa->n = n;
    asa->ns = ns;
    asa->t = t;
    asa->c = c;
    asa->aap = aap;  
    asa->send = send_asa;
    asa->delete = delete;
  
    asa->mseq = 0;
    asa->rseq = rseq;
    asa->last_exp_tim = 0;


    return asa;
  

}
