#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "aap_aclm.h"
#include "defender.h"

/*------------------------------------------------------------------*/
/*
 * defense_timeout_callback
 *
 */
/* from draft-ietf-malloc-aap-03.txt:
   
   If MAAS A's Allocation Defense timer expires, it SHOULD send an AIU
   message containing the addresses in question and restart the timer
   with twice its previous value, unless that value was 0, in which case
   it SHOULD set it to [RESEND-WAIT]. If the timer interval becomes
   greater than [REPEAT-INTERVAL], the timer SHOULD be cancelled. In
   this case, the remote MAAS probably suffered some form of failure
   after sending the ACLM or AITU.
*/
static int defense_timeout_callback(void *obj, unsigned int time) {
    int pos;
    defender *def = obj;
    aap_aiu *aiu = def->aiu;
    /*  allocation *alloc = def->alloc;*/
    /*  addr_range *ar = alloc->range(alloc);*/

    printf("Defense timeout:\n");

    aiu->send(aiu);
    if(def->last_interval == 0) {
        def->last_interval = def->cfg->RESEND_WAIT;
    } else {
        def->last_interval = def->last_interval * 2;
    }
    printf("last interval %d\n", def->last_interval);
    if(def->last_interval >= def->cfg->REPEAT_INTERVAL) {
        def->cancel(def);
        pos = def->aap->defens_list->find(def->aap->defens_list, def);
        def->aap->defens_list->remove(def->aap->defens_list, pos);
        def->delete(&def);
    } else {
        def->defense_timer = def->t->set(def->t, def->last_interval, defense_timeout_callback, def); 
    }
  
    return 0;
}

/*------------------------------------------------------------------*/
/* from draft-ietf-malloc-aap-03.txt:

   If MAAS A sees an AIU message from someone other than MAAS B
   containing one or more of the addresses in question before the
   Allocation Defense timer expires, it SHOULD restart the timer with
   double its original value. If MAAS A sees an ACLM or AITU message
   from MAAS B with the same Request Sequence Number and a different set
   of addresses, it SHOULD cancel the Allocation Defense timer and check
   this message against the Allocation Record as described above.
*/

static void defend(defender *def) {
    int resend_wait;
    allocation *alloc = def->alloc;

    def->addr_list = list_new();
    def->addr_list->insert(def->addr_list, alloc);
    def->aiu = aap_aiu_new(def->n, def->ns, def->t, def->cfg, def->aap,
                           def->addr_list, def->aap->next_rseq(def->aap));
    
    if(alloc->owner_addr(alloc) == ALLOCATION_OWN) {
        resend_wait = 0;
    } else {
        resend_wait = ((double)(rand())/RAND_MAX) * (def->cfg->RESEND_WAIT*6) +
            (def->cfg->RESEND_WAIT*2);	  	  
    }
  
    def->defense_timer = def->t->set(def->t, resend_wait,
                                     defense_timeout_callback, def);
    def->last_interval = resend_wait;

}
/*------------------------------------------------------------------*/
static void cancel(defender *def) {
    printf("Cancelling defens\n");
    def->t->unset(def->t, def->defense_timer);
}
/*------------------------------------------------------------------*/
static void delete(defender **def) {
    (*def)->aiu->delete((*def)->aiu);
    (*def)->addr_list->delete((*def)->addr_list);
    free(*def);
    *def = NULL;
}
/*------------------------------------------------------------------*/
static void double_interval(defender *def) {
    def->t->unset(def->t, def->defense_timer);
    def->last_interval *= 2;
    def->defense_timer = def->t->set(def->t, def->last_interval, defense_timeout_callback, def);
  
}
/*------------------------------------------------------------------*/
static addr_range *range(defender *def) {
    return def->alloc->range(def->alloc);
}
/*------------------------------------------------------------------*/
static int match(defender *def, int rseq, unsigned long sender) {
    return (def->rseq == rseq && def->sender == sender);	  
}
/*------------------------------------------------------------------*/
static int overlap(defender *def, addr_range *ar) {
    return ar->overlap(ar, def->alloc->range(def->alloc));
}
/*------------------------------------------------------------------*/
defender *defender_new(aap *aap, allocation *alloc, int rseq, unsigned long sender) {
    defender *def = malloc(sizeof(defender));

    def->n = aap->net;
    def->ns = aap->out_ns;
    def->t = aap->tim;
    def->cfg = aap->cfg;
    def->aap = aap;
    def->alloc = alloc;

    def->rseq = rseq;
    def->sender = sender;
  
    def->defend = defend;
    def->range  = range;
    def->cancel = cancel;
    def->delete  = delete;
    def->match  = match;
    def->overlap  = overlap;
    def->double_interval = double_interval;
    return def;
}
