/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains the main module for the MAAS (Multicast Address
 * Allocation Server). It configurates and starts the AAP and MADCAP protocols.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

/* TODO: In AAP some code still excists to handle a simple
   testclient (client_protocol) and also some code that makes AAP a
   simple Prefix Coordinator. This code does not fulfill anything and
   can be removed. */
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
/* MADCAP */
#include "madcap.h"
#include "madcap_list.h"
#include "debug.h"
/* AAP */
#include "tim.h"
#include "net.h"
#include "aap.h"
#include "client_protocol.h"


static void usage() {
    printf("Usage: maasd [-v verbose_options] [-a aap_config_file] [-m madcap_config_file] aap_config_name1 [aap_config_name2 ...] [madcap_config_name]\n");
    exit(1);
} 

/**
 * This is the main function for the MAAS server. It is responsible for
 * creating and starting the aap's and the madcap.
 * @param argc The number of arguments supplied to the program.
 * @param argv Contains pointers to strings representing the arguments
 * given at startup.
 * @return An integer.
 */
int main(int argc, char **argv){
    /* Common */
    tim *timer;
    net *net;    
    /* MADCAP */
    madcap *madcap;
    madcap_cfg *mc_cfg;
    madcap_error *mc_error;
    madcap_list *aap_list;
    madcap_list *mc_cfg_list;
    char *mc_configfname = NULL;
    aap_list_item *ali;
    int mc_configured = 0; /* Used to check that only one MADCAP configuration
                              is given as argument. */
    /* AAP */
    aap *aap;
    cfg *aap_cfg;
    uint32 scope_id;
    struct in_addr scope_addr;
    list *cfg_list;
    int i, j;
    list *cfgnames;
    char *configfname = NULL;
    
    /* Initiating the MAAS */
    timer = tim_new();
    net = net_new(timer);
    /* MADCAP */
    aap_list = madcap_list_new();
    mc_error = madcap_error_new();
    if (mc_error == NULL) {
        ABORT("Initiation of MADCAP_ERROR failed.");
    }

    /* Get arguments. */
    for(j = 1; j < argc && argv[j][0] == '-'; j++) {
        switch(argv[j][1]) {
        case 'v': /* Verbose mode. */
            /* Check that there exist an argument after the flag. */
            if(j + 1 >= argc) {
                usage();
            }
            DEBUG_INIT(argv[j + 1]);
            printf("Verbose mode enabled with flag '%s'.\n", argv[j + 1]);
            j++;
            break;
        case 'a': /* AAP configuration file. */
            if(j + 1 >= argc) {
                usage();
            }
            configfname = argv[j + 1];
            if(access(configfname, R_OK) == -1) {
                fprintf(stderr, "Configuration file %s unreadable\n",
                        configfname);
                exit(1);
            }
            j++;
            break;
        case 'm': /* MADCAP configuration file. */
            if(j + 1 >= argc) {
                usage();
            }
            mc_configfname = argv[j + 1];
            if(access(mc_configfname, R_OK) == -1) {
                fprintf(stderr, "Configuration file %s unreadable\n",
                        mc_configfname);
                exit(1);
            }
            j++;
            break;
        default:
            usage();
        }
    }

    /* Check that at least one configuration have been chosen. */
    if(j == argc) {
        usage();
    }

    /* Get the unique configuration names for AAP and MADCAP. */
    cfgnames = list_new();
    for(;j < argc && argv[j][0] != '-'; j++) {
        cfgnames->insert(cfgnames, argv[j]);
    }
   
    if(configfname == NULL) {
        /* No AAP configuration filename given as argument. */
        if(access("/etc/aap.conf", R_OK) == 0) {
            configfname = "/etc/aap.conf";    
        } else if(access("/usr/local/etc/aap.conf", R_OK) == 0) {
            configfname = "/usr/local/etc/aap.conf";
        }
    
        if(configfname == NULL) {
            fprintf(stderr, "No configuration file /etc/aap.conf or /usr/local/etc/aap.conf found.\n");
            exit(1);
        }
    }
    
    if(mc_configfname == NULL) {
        /* No MADCAP configuration filename given as argument. */
        if(access("/etc/madcap.conf", R_OK) == 0) {
            mc_configfname = "/etc/madcap.conf";
        } else if(access("/usr/local/etc/madcap.conf", R_OK) == 0) {
            mc_configfname = "/usr/local/etc/madcap.conf";
        }
    }
      
    if(mc_configfname == NULL) {
        fprintf(stderr,"No configuration file /etc/madcap.conf or ");
        fprintf(stderr, "/usr/local/etc/madcap.conf found.\n");
        exit(1);
    }
      
    /* Reading AAP configuration file. */
    cfg_list = cfg_parse(configfname);
    if(DEBUG_ENABLED('i') == 0) {
        printf("AAP file contains %d configurations:",
              (cfg_list->length(cfg_list)));
        for(i = 0; i < cfg_list->length(cfg_list); i++) {
            aap_cfg = (cfg *)cfg_list->elem_at(cfg_list, i);
            printf(" %s",aap_cfg->name);
        }
        printf(".\n");
    }
      
    for(i = 0; i < cfg_list->length(cfg_list); i++) {
        aap_cfg = (cfg *)cfg_list->elem_at(cfg_list, i);
        for(j = 0; j < cfgnames->length(cfgnames); j++) {
            if(strcmp(cfgnames->elem_at(cfgnames, j), aap_cfg->name) == 0) {
                client_protocol *cp;
                printf("Starting AAP configuration '%s'\n", aap_cfg->name);
                aap = aap_new(net, timer, aap_cfg);
                aap->start(aap);

                /* Store scope ID of this AAP in list to MADCAP. */
                scope_id = aap->cfg->scope->first_addr(aap->cfg->scope);
                scope_addr.s_addr = scope_id;
                DEBUG('i', "Scope ID: %s\n", inet_ntoa(scope_addr));
                ali = malloc(sizeof(aap_list_item));
                ASSERT(ali != NULL, "Memory allocation failed.");
                ali->aap = aap;
                ali->scope_id = ntohl(scope_id);
                aap_list->insert(aap_list, ali);

                /* Enable communication with old AAP test client. */
                cp = client_protocol_new(net, timer, aap);
                cp->start(cp);
	
                break;
            } else {
                /* TODO: This should be done if the configuration files
                   consists of alot of defined configurations to avoid
                   unneccessary usage of memory. Though, it requires a rewrite
                   of these for-loops. */
                /*aap_cfg->delete(aap_cfg);*/
            }
        }
    }
    cfg_list->delete(cfg_list);


    /* Reading MADCAP configuration file. */
    mc_cfg_list = mc_cfg_parse(mc_configfname);
    if(DEBUG_ENABLED('i') == 0) {
        DEBUG('i',
            "MADCAP file contains %d configurations:",
          (mc_cfg_list->length(mc_cfg_list)));
        for(i = 0; i < mc_cfg_list->length(mc_cfg_list); i++) {
            mc_cfg = (madcap_cfg *)mc_cfg_list->elem_at(mc_cfg_list, i);
            printf(" %s",mc_cfg->name);
        }
        printf(".\n");
    }
      
    for(i = 0; i < mc_cfg_list->length(mc_cfg_list); i++) {
        mc_cfg = (madcap_cfg *)mc_cfg_list->elem_at(mc_cfg_list, i);
        for(j = 0; j < cfgnames->length(cfgnames); j++) {
            if(strcmp(cfgnames->elem_at(cfgnames, j), mc_cfg->name) == 0) {
                printf("Starting MADCAP configuration '%s'\n",
                      mc_cfg->name);
                if(mc_configured == 1){
                    fprintf(stderr,
                        "Failure: only one MADCAP configuration is allowed.\n");
                    usage();
                } else {
                    mc_configured = 1;
                }
                madcap = madcap_new(net, timer, mc_cfg, aap_list);
                madcap->start(madcap);
                break;
            } else {
                /* TODO: This should be done if the configuration files
                   consists of alot of defined configurations to avoid
                   unneccessary usage of memory. Though, it requires a rewrite
                   of these for-loops. */
                /*mc_cfg->delete(mc_cfg);*/
            }
        }
    }
    if(mc_configured == 0) {
        printf("Starting default MADCAP configuration.\n");
    }
      
    /* Unleash MAAS */
    printf("\nStarting MAAS...\n");
    net->main(net); /* No return from this call. */
    return 0;
}
