/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file is the top-level implementation of the MADCAP part of MAAS.
 * The file instansiate the network socket by which client connects to.
 * The MADACP object also implements the receiving function that is
 * called whenever data is arriving from the network.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */
#include "madcap.h"


#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include "madcap_msg_nak.h"
#include "madcap_msg_ack.h"
#include "madcap_msg_offer.h"

#include "madcap_msg_request.h"
#include "madcap_list.h"
#include "madcap_cfg.h"
#include "madcap_msg.h"
#include "madcap_msg_discover.h"
#include "madcap_msg_getinfo.h"
#include "madcap_msg_release.h"
#include "madcap_msg_renew.h"
#include "madcap_msg_unknown.h"
#include "madcap_error.h"
#include "addr_rec.h"
#include "net_socket.h"
#include "net.h"

static int pub_start(madcap *object);
static int pub_save_outgoing_msg(madcap *mc, uint32 xid,
                                 madcap_msg_type msg_type,
                                 madcap_msg_type resonse_type,
                                 uint8 *lease_id,
                                 uint16 lease_id_len,
                                 void *msg);
static int receive_msg(void *object, char *data, int len, net_peer *np);
static int response_cache_timeout(void *object);

/**
 * Creates a new MADCAP object.
 *
 * @param net A network abstraction object.
 * @param tim An abstract timer object.
 * @param mc_cfg The MADCAP configuration.
 * @param aap_list The list with aap's.
 * @return The newly created object. A return value of NULL
 * indicates a failure to create the new object.
 */
madcap *madcap_new(net *net, tim *tim, madcap_cfg *mc_cfg,
                   madcap_list *aap_list){
    char filename[50];
    madcap_list *cache_list = NULL;
    /* Create the object and assign data to the variables in the struct. */
    madcap *object = (madcap *)malloc(sizeof(madcap));    
    ASSERT(object != NULL, "Memory allocation failed.");

    
    object->net               = net;
    object->tim               = tim;
    object->mc_cfg            = mc_cfg;    
    object->start             = pub_start;
    object->save_outgoing_msg = pub_save_outgoing_msg;
    
    
    /* Create incoming and outgoing sockets. */
    object->in_socket  = net->socket(net);
    object->out_socket = net->socket(net);

    /* Try to read from the cache. */
    mc_cfg->get_addr_rec_file(mc_cfg, filename);
    if(access(filename, R_OK) == 0) {
        cache_list = mc_cfg_parse(filename);
    }

    /* Create a new address record. */
    object->a_rec = addr_rec_new(mc_cfg, tim, aap_list, cache_list);

    /* Create a new priority list for sent messages. */
    object->response_cache = madcap_plist_new();

    return object;    
}

/*
 * Private method to match incoming duplicate messages.
 */
static int match(void *obj, void *element){
    msg_id *find = obj;
    msg_id *list_elem = element;
    int retval;
    
    if((find->xid == list_elem->xid) && (find->msg_type == list_elem->msg_type)
       && (find->lease_len == list_elem->lease_len)) {
        if(memcmp(find->lease_id, list_elem->lease_id, find->lease_len) == 0) {
            retval = 1; /* Matching. */
        }
    } else {
        retval = 0; /* Not matching. */
    }
    return retval;
}

/**
 * Function called when data arrives from the network.
 *
 * @param object The object passed as argument when this receive
 * function was attached as receive function to the net object.
 * @param data The arriving data.
 * @param len The length of the data in bytes.
 * @param np The address of the sender of the data.
 * @return -1 on failure, otherwise 0.
 */
static int receive_msg(void *object, char *data, int len, net_peer *np){
    madcap_msg *incoming_msg          = madcap_msg_new();
    int ret_val                       = 0;
    int response_cache_interval       = 0;
    madcap_msg_getinfo *msg_getinfo   = NULL;
    madcap_msg_request *msg_request   = NULL;
    madcap_msg_release *msg_release   = NULL;
    madcap_msg_unknown *msg_unknown   = NULL;
    madcap_msg_discover *msg_discover = NULL;
    madcap_msg_nak *tmp_nak           = NULL;
    madcap_msg_ack *tmp_ack           = NULL;
    madcap_msg_offer *tmp_offer       = NULL;
    madcap_msg_renew *msg_renew       = NULL;        
    madcap_msg_type type              = 0;
    madcap_error_type error           = 0;
    madcap *self                      = (madcap *)object;
    msg_id *incoming_msg_id           = (msg_id *)malloc(sizeof(msg_id));
    msg_id *former_msg_id             = NULL;
    
    if (incoming_msg == NULL) {
        return -1;
    }    

    /* Set destination to the current client. */
    self->out_socket->set_dest(self->out_socket,
                               self->net->sender->ip_addr,
                               self->net->sender->peer_port);
    
    /* Parse incoming data. */
    ret_val = incoming_msg->parse(incoming_msg, data, len, np);

    if(ret_val == -1) {
        /* Get cause. */
        mc_error->madcap_errnum_collect(mc_error, &error);
        /* Set back the same cause(since we collected it)*/
        mc_error->notify(mc_error, error, "");
        
        switch(error) {
        case MC_ERR_MSGLEN:
            /* Ignore according to section 2.6 in rfc2730. */
            /* Delete incoming message. */
            incoming_msg->delete(incoming_msg);
            return -1;
            break;
            
        case MC_ERR_VERSION:
            /* Ignore according to section 2.6 in rfc2730. */
            /* Delete incoming message. */
            incoming_msg->delete(incoming_msg);
            return -1;
            break;
            
        case MC_ERR_MSGTYPE:
            /* Respond according to section 2.2 in rfc2730. */
            msg_unknown =  madcap_msg_unknown_new(incoming_msg,
                                                  self->net,
                                                  self->out_socket,
                                                  MC_ERR_MSGTYPE);
            if (msg_unknown == NULL) {
                return -1;
            }

            ret_val = msg_unknown->receive(msg_unknown);             
            msg_unknown->delete(msg_unknown);
            /* Delete incoming message. */
            incoming_msg->delete(incoming_msg);
            return -1;        
            break;
            
        case MC_ERR_FAMILY:
            /* Ignore according to section 2.6 in rfc2730. */
            /* Delete incoming message. */
            incoming_msg->delete(incoming_msg);
            return -1;
            break;
            
        case MC_ERR_OPTIONS:
            /* Respond according to section 2.2 in rfc2730. */
            msg_unknown =  madcap_msg_unknown_new(incoming_msg,
                                                  self->net,
                                                  self->out_socket,
                                                  MC_ERR_OPTIONS);
            if (msg_unknown == NULL) {
                return -1;
            }
            ret_val = msg_unknown->receive(msg_unknown);
            msg_unknown->delete(msg_unknown);
            /* Delete incoming message. */
            incoming_msg->delete(incoming_msg);
            return -1;
            break;
            
        default:
            /* Unknown error */
            /* Delete incoming message. */
            incoming_msg->delete(incoming_msg);
            return -1;
        }        
    }
        
    ret_val = incoming_msg->get_msg_type(incoming_msg, &type); 
    if(ret_val == -1) {
        /* Respond according to section 2.2 in rfc2730. */
        msg_unknown =  madcap_msg_unknown_new(incoming_msg,
                                              self->net,
                                              self->out_socket,
                                              MC_ERR_MSGTYPE);
        if (msg_unknown == NULL) {
            return -1;
        }
        ret_val = msg_unknown->receive(msg_unknown);
        msg_unknown->delete(msg_unknown);
        incoming_msg->delete(incoming_msg);
        return -1;
    }
    
    /* Check if this message already have been reveived and responded.
     * If responded to earlier within [RESPONSE_CACHE_INTERVAL], retransmit
     * respons message.
     *
     * Messages is cached in the server right before sending, and therefore
     * ignored messages is processed again. Performance gains may be achieved
     * by storing ignored messages in next iteration of this implementation.
     */
    incoming_msg->get_xid(incoming_msg, &incoming_msg_id->xid);
    incoming_msg->get_msg_type(incoming_msg, &incoming_msg_id->msg_type);
    ret_val = incoming_msg->get_option(incoming_msg,
                                       MC_LEASE_ID,
                                       &incoming_msg_id->lease_len,
                                       &incoming_msg_id->lease_id);
    if(ret_val == SUCCESS) {
        former_msg_id = NULL;
        /* Check if dublicate message */
        former_msg_id =
            (msg_id *)self->response_cache->find_element(self->response_cache,
                                                         match,
                                                         incoming_msg_id);

        /* Dublicate message arrived */
        if(former_msg_id != NULL) {
            switch(former_msg_id->response_type) {
            case MC_NAK:
                DEBUG('m',"Dublicate NAK message received.\n");
                tmp_nak = (madcap_msg_nak *)former_msg_id->sent_msg;
                tmp_nak->send(tmp_nak);                
                break;

            case MC_ACK:
                DEBUG('m',"Dublicate ACK message received.\n");
                tmp_ack = (madcap_msg_ack *)former_msg_id->sent_msg;
                tmp_ack->send(tmp_ack);               
                break;

            case MC_OFFER:
                DEBUG('m',"Dublicate OFFER message received.\n");
                tmp_offer = (madcap_msg_offer *)former_msg_id->sent_msg;
                tmp_offer->send(tmp_offer);               
                break;
            default:
            }

            /* Removes the message from the list. It will be inserted in
             * the list again, but with a new time value
             * of [RESPONSE_CACHE_INTERVAL].
             */ 
            self->tim->unset(self->tim, former_msg_id->timer);
            self->response_cache->remove_element(self->response_cache,
                                                 match,
                                                 former_msg_id);
            
            self->mc_cfg->get_response_cache_interval(self->mc_cfg,
                                                      NULL,
                                                      &response_cache_interval);
            /* Set new timer for deletion for this sent message */
            former_msg_id->timer =
                self->tim->set(self->tim,
                               response_cache_interval,
                               (tim_callback)response_cache_timeout,
                               self);            
            
            /* Insert the message again with a new timevalue,
             * since it might not reach the client once again
             */
            self->response_cache->insert(self->response_cache,
                                         (uint32)time(NULL),
                                         former_msg_id);

            
        } else {
            /* New messafe arrived, check what type of incoming
             * message it was, create then corresponding message
             * to process the request from the client.
             */
            ret_val = 0;
            switch(type) {
            case MC_REQUEST:
                DEBUG('m',"REQUEST message received from %s.\n",
                      self->net->sender->ip_addr);
                msg_request = madcap_msg_request_new(incoming_msg,
                                                     self->mc_cfg,
                                                     self->a_rec,
                                                     self->net,
                                                     self->out_socket,
                                                     self);
                if (msg_request == NULL) {
                    return  -1;
                }
                ret_val = msg_request->receive(msg_request);
                msg_request->delete(msg_request);
                break;
        
            case MC_RELEASE:
                DEBUG('m',"RELEASE message received from %s.\n",
                      self->net->sender->ip_addr);
                msg_release = madcap_msg_release_new(incoming_msg,
                                                     self->a_rec,
                                                     self->net,
                                                     self->out_socket,
                                                     self);
                if (msg_release == NULL) {
                    return -1;
                }
                ret_val = msg_release->receive(msg_release);
                msg_release->delete(msg_release);
                break;
        
            case MC_GETINFO:
                DEBUG('m',"GETINFO message received from %s.\n",
                      self->net->sender->ip_addr);
                msg_getinfo = madcap_msg_getinfo_new(incoming_msg,
                                                     self->net,
                                                     self->out_socket,
                                                     self);
                if (msg_getinfo == NULL) {
                    return -1;
                }
                ret_val = msg_getinfo->receive(msg_getinfo);
                msg_getinfo->delete(msg_getinfo);                        
                break;
        
            case MC_DISCOVER:
                DEBUG('m',"DISCOVER message received from %s.\n",
                      self->net->sender->ip_addr);
                msg_discover = madcap_msg_discover_new(incoming_msg,
                                                       self->mc_cfg,
                                                       self->a_rec,
                                                       self->net,
                                                       self->out_socket,
                                                       self);
                if (msg_discover == NULL) {
                    return -1;
                }
                ret_val = msg_discover->receive(msg_discover);
                msg_discover->delete(msg_discover);            
                break;
           
            case MC_RENEW:
                DEBUG('m',"RENEW message received from %s.\n",
                      self->net->sender->ip_addr);
                msg_renew = madcap_msg_renew_new(incoming_msg,
                                                 self->mc_cfg,
                                                 self->a_rec,
                                                 self->net,
                                                 self->out_socket,
                                                 self);
                if (msg_renew == NULL) {
                    return -1;
                }
                ret_val = msg_renew->receive(msg_renew);
                msg_renew->delete(msg_renew);            
                break;
        
            case MC_NAK:
                /*
                 * Do nothing to avoid NAK loops according to
                 * section 2.6 in RFC2730.
                 */
                break;
            default:
                /* Erranous message received. */
                /*
                 * Generate invalid request error according to
                 * section 2.2 and 2.6 in the RFC2730 and respond with a NAK.
                 */
                DEBUG('m',"Unknown message type received from %s.\n",
                      self->net->sender->ip_addr);
                msg_unknown =  madcap_msg_unknown_new(incoming_msg,
                                                      self->net,
                                                      self->out_socket,
                                                      MC_ERR_MSGTYPE);
                if (msg_unknown == NULL) {
                    return -1;        
                }
                ret_val = msg_unknown->receive(msg_unknown);
                msg_unknown->delete(msg_unknown);
            }
        }
    }

    
    /* Delete incoming message. */
    incoming_msg->delete(incoming_msg);
    free(incoming_msg_id);
    DEBUG('m', "Madcap: Incoming message processed.\n");  
    return ret_val;
}


/**
 * Function called to start the MADCAP part of the MAAS.
 * Reads the current configuration and sets up the networking
 * part to communicate with the clients.
 * @param object The MADCAP object itself.
 * @return -1 on failure, otherwise 0.
 */   
static int pub_start(madcap *object){
    int number_scopes       = 0;
    int i                   = 0;
    int port                = 0;    
    int ret_val             = 0;
    uint32 scope_addr       = 0;
    unsigned int ttl        = 0;
    scope *scps;
    
    /* Read the configuration for scopes and portnumber*/
    ret_val = object->mc_cfg->get_number_of_scopes(object->mc_cfg,
                                                   &number_scopes);
    ASSERT(ret_val != -1, "Could not get number of scopes from configuration.");
    
    /* Get port. */
    ret_val = object->mc_cfg->get_port(object->mc_cfg,
                                       object->mc_cfg->name,
                                       &port);
    ASSERT(ret_val != -1, "Could not get port from configuration.");
    
    /* Bind incomming socket */    
    ret_val = object->in_socket->bind(object->in_socket,
                                      NET_SOCKET_ANY,
                                      port);    
    ASSERT(ret_val != -1, "Could not bind port.");    

    /* Setting callback function(named 'receive_msg')
     * to be called when messages arrives at socket.
     */
    object->net->receive(object->net, object->in_socket,
                         receive_msg, object);
    
    /* Get all scopes, and join corresponding multicast groups */
    ret_val = object->mc_cfg->get_scopes(object->mc_cfg, &scps);
    ASSERT(ret_val != -1, "Could not get the scopes.");    

    while(i < number_scopes) {
        scope_addr = (scps[i]).addresses.end - 1;
        ttl        = (scps[i]).ttl;
        ret_val = object->in_socket->join_multicast_group(object->in_socket,
                                                           scope_addr,
                                                           ttl);
        ASSERT(ret_val != -1, "Could not join multicast group.");
        i++;
    } 
        
    return 0;
}

/**
 * Function called to save the last outgoing message(ACK,NAK,OFFER).
 * Needed since if a dublicate message is received within
 * [RESPONSE_CACHE_INTERVAL], a identical response is sent to the client.
 * @param mc The MADCAP object itself. 
 * @param xid The transaction id of the message.
 * @param msg_type The message type of the last incoming message.
 * @param response_type The message type of the last message sent. 
 * @param lease_id The lease identifier in the last message.
 * @param lease_id_len The length of the lease identifier.
 * @param msg The message sent.
 * @return -1 on failure, otherwise 0. 
 */ 
static int pub_save_outgoing_msg(madcap *mc, uint32 xid,
                                 madcap_msg_type msg_type,
                                 madcap_msg_type response_type,
                                 uint8 *lease_id,
                                 uint16 lease_id_len,
                                 void *msg){

    msg_id *sent_msg            = (msg_id *)malloc(sizeof(msg_id));
    int response_cache_interval = 0;
    int ret_val                 = 0;
    sent_msg->xid               = xid;
    sent_msg->msg_type          = msg_type;
    sent_msg->response_type     = response_type;
    sent_msg->lease_id          = lease_id;
    sent_msg->lease_len         = lease_id_len;
    sent_msg->sent_msg          = msg;

    /* Get [RESPONSE_CACHE_INTERVAL] from configuration. */
    ret_val = mc->mc_cfg->get_response_cache_interval(mc->mc_cfg,
                                                      mc->mc_cfg->name,
                                                      &response_cache_interval);
    ASSERT(ret_val != -1, "Could not get [RESPONSE_CACHE_INTERVAL].");

    /* Set time ([RESPONSE_CACHE_INTERVAL] seconds) when to delete
     * this message. */
    sent_msg->timer = mc->tim->set(mc->tim,
                                   response_cache_interval,
                                   (tim_callback)response_cache_timeout,
                                   mc);

    /* Store the message in the 'response_cache' list. */
    mc->response_cache->insert(mc->response_cache,
                               (uint32)time(NULL),
                               sent_msg);
    return 0;
}

/* Callback function called when [RESPONSE_CACHE_INTERVAL] timeouts.
 * @param object The MADCAP object itself.
 */
static int response_cache_timeout(void *object){
    madcap_msg_nak *tmp_nak     = NULL;
    madcap_msg_ack *tmp_ack     = NULL;
    madcap_msg_offer *tmp_offer = NULL;
    madcap *self                = (madcap *)object;

    /* Get first message from list. Since it is a priority list,
     * (with time as priority), the first element in the list is the one to
     * delete from the list.
     */
    msg_id *tmp = self->response_cache->first(self->response_cache);
    
    if (tmp != NULL) {
        self->response_cache->remove(self->response_cache, 0);
    } else {
        return -1;
    }
    /* Free memory of old message. */
    switch(tmp->response_type) {
    case MC_NAK:
        DEBUG('m',"NAK message removed from cache\n"); 
        tmp_nak = (madcap_msg_nak *)tmp->sent_msg;
        tmp_nak->delete(tmp_nak);
        break;
        
    case MC_ACK:
        DEBUG('m',"ACK message removed from cache\n"); 
        tmp_ack = (madcap_msg_ack *)tmp->sent_msg;
        tmp_ack->delete(tmp_ack);
        break;
        
    case MC_OFFER:
        DEBUG('m',"OFFER message removed from cache\n"); 
        tmp_offer = (madcap_msg_offer *)tmp->sent_msg;
        tmp_offer->delete(tmp_offer);
    default:
        
    }
    free(tmp);    
    return 0;
}
