/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains constructor and methods for a madcap_error object.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "madcap_error.h"

/* Default error-messages. */
static char *err_msg[] = {
    "Bad address family",
    "Wrong version of MADCAP message",
    "Bad message type",
    "Unknown transaction ID",
    "Unknown lease ID",
    "Length of options field too long",
    "Bad syntax in options field",
    "Length of MADCAP message too short",
    "Bad scope",
    "Bad desired number of addresses",
    "Bad desired lifetimes",
    "Bad start time durations",
    "Requested start times conflict with requested lifetimes",
    "The operation failed",
    "The addresses were not the ones that had been allocated",
    "Nesting state undetermined at this time",
    "Not enough addresses available",
    "Not enough lifetime on available addresses",
    "Could not create DISCOVER message",
    "Could not create OFFER message",
    "Could not create REQUEST message",
    "Could not create RENEW message",
    "Could not create ACK message",
    "Could not create NAK message",
    "Could not create RELEASE message",
    "Could not create GETINFO message",
    "Could not send OFFER message",
    "Could not send ACK message",
    "Could not send NAK message",
    "Type of error unknown",
    "The type of error is not specified",
    "The error is already collected, and no new errors has occured",
    "Configuration Error"
};

/**
 * Public method to notify which error has ocurred.
 * 
 * @param object The madcap_error-object.
 * @param err_num The error number.
 * @param message The error message.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_notify(madcap_error *object, madcap_error_type err_num,
                      char *message){
    object->mc_errno = err_num;
    object->err_message = message;
    return SUCCESS;
}

/**
 * Public method to collect the error message as a describing string.
 * 
 * @param object The madcap_error-object.
 * @param msg The error message.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_madcap_errstring_collect(madcap_error *object, char *msg){
    
    char *default_error_str;
    int ret_val;
    if(object->err_message != NULL) {
        default_error_str = (char *)malloc(100 * sizeof(char));
        if(default_error_str == NULL) {
            fprintf(stderr, "Could not allocate memory\n");
            return FAILURE;
        }
        ret_val = object->madcap_errstring_translate(object, object->mc_errno,
                                                     default_error_str);
        if(ret_val == FAILURE) {
            return FAILURE;
        }
        strcpy(msg, default_error_str);
        strcat(msg, ": ");
        strcat(msg, object->err_message);
        free(default_error_str);
    } else {
        ret_val = object->madcap_errstring_translate(object,
                                                     object->mc_errno, msg);
        if(ret_val == FAILURE) {
            return FAILURE;
        }
    }
    object->mc_errno = MC_ERR_NONEW;
    object->err_message = NULL;
    return SUCCESS;
}

/**
 * Public method to collect the error message as a number.
 *
 * @param object The madcap_error-object.
 * @param err_num The error number.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_madcap_errnum_collect(madcap_error *object,
                                     madcap_error_type *err_num){
    *err_num = object->mc_errno;
    object->mc_errno = MC_ERR_NONEW;
    object->err_message = NULL;
    return SUCCESS;
}

/**
 * Public method to translates an error type from number to a string.
 *
 * @param object The madcap_error-object.
 * @param err_num The error type number.
 * @param error A string with an error message describing the error.
 * @return A return value of 0 indicates a success and a return
 * value of -1 indicates a failure. 
 */
static int pub_madcap_errstring_translate(madcap_error *object,
                                          madcap_error_type err_num,
                                          char *error){
    if(err_num < MC_ERR_MAX) {
        strcpy(error, err_msg[err_num]);
        return SUCCESS;
    }
    strcpy( error,  "Unknown error type, system failure.");
    return FAILURE;
}

/**
 * Pulbic method that prints the description of the error
 * to standard error output.
 * If message is NULL, the function prints the error-message held by
 * the object. If message is not NULL, the function prints the
 * message, a colon, a space, and then the error-message.
 *
 * @param object The madcap_error-object.
 * @param message The message to append to the default error message.
 * @return A return value of 0 indicates a success and a return value
 * of -1 indicates a failure.
 */
static int pub_madcap_perror(madcap_error *object, const char *message){
    char *default_error_msg;
    int ret_val;
    default_error_msg = (char *)malloc(100 * sizeof(char));
    if(default_error_msg == NULL) {
        return FAILURE;
    }
    ret_val = object->madcap_errstring_translate(object, object->mc_errno,
                                                 default_error_msg);
    if(ret_val == FAILURE) {
        return FAILURE;
    }
    if(message == NULL) { 
        fprintf(stderr, "%s\n", default_error_msg);
    } else {
        fprintf(stderr, "%s: %s\n", message, default_error_msg);
    }
    free(default_error_msg);
    return SUCCESS;
}

/**
 * The constructor of a madcap_error object.
 *
 * @return A pointer to a madcap_error-object on succes, a NULL-pointer in
 * case of failure.
 */
madcap_error *madcap_error_new(){
    madcap_error *object;
    object = (madcap_error *)calloc(1, sizeof(madcap_error));
    if(object == NULL) {
        fprintf(stderr,
                "Could not allocate enough memory for madcap_error-struct\n");
        return NULL;
    }
    /* Set the function-pointers. */
    object->notify = pub_notify;
    object->madcap_errstring_collect = pub_madcap_errstring_collect;
    object->madcap_errnum_collect = pub_madcap_errnum_collect;
    object->madcap_errstring_translate = pub_madcap_errstring_translate;
    object->madcap_perror = pub_madcap_perror;
    mc_error = object;    
    return object;
}
