/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains the structure and definitions for a generic MADCAP
 * message.
 *
 * The data fields, as defined in RFC2730 December 1999, are:
 *
 * FIELD      OCTETS  DESCRIPTION
 * -----      ------  -----------
 * version      1     Protocol version number (zero for this specification)
 * msgtype      1     Massage type (DISCOVER, GETINFO, etc.)
 * addrfamily   2     Address family (IPv4, IPv6, etc.)
 * xid          4     Transaction ID
 * options     var    Options field
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */

#ifndef _MADCAP_MSG_H_
#define _MADCAP_MSG_H_

#include "globals.h"
#include "madcap_error.h"
#include "net.h"
#include "net_socket.h"

/* The IPv4 address family as defined by the IANA. */
#define MC_ADDR_FAMILY 1

/* Minimum length of a MADCAP message. */
#define MIN_MSG_LENGTH 12

/* Header size of a MADCAP message. */
#define MADCAP_HEADER 8

/* Version of MADCAP message. */
#define MADCAP_VERSION 0
  
/* Maximum size in bytes of the options field in a MADCAP message. */
#define MAX_OPTIONS 2048

typedef struct madcap_msg_ madcap_msg;

/* The different types of MADCAP massages. */
enum madcap_msg_type {
    MC_DISCOVER = 1,
    MC_OFFER    = 2,
    MC_REQUEST  = 3,
    MC_RENEW    = 4,
    MC_ACK      = 5,
    MC_NAK      = 6,
    MC_RELEASE  = 7,
    MC_GETINFO  = 8,
    MC_MSGTYPE_VALID = 9 /* Limit to use when checking if type is valid. */
};
typedef enum madcap_msg_type madcap_msg_type;

/* The different types of options in a MADCAP message. */
enum madcap_opt_type {
    MC_END             = 0,
    MC_LEASE_TIME      = 1,
    MC_SERVER_ID       = 2,
    MC_LEASE_ID        = 3,
    MC_SCOPE           = 4,
    MC_REQUEST_LIST    = 5,
    MC_START_TIME      = 6,
    MC_NO_ADDR_REQ     = 7,
    MC_REQ_LANGUAGE    = 8,
    MC_SCOPE_LIST      = 9,
    MC_ADDR_RANGE_LIST = 10,
    MC_CURRENT_TIME    = 11,
    MC_FEATURE_LIST    = 12,
    MC_RETRY_TIME      = 13,
    MC_MIN_LEASE_TIME  = 14,
    MC_MAX_START_TIME  = 15,
    MC_ERROR           = 16,
    MC_OPT_VALID       = 17 /*Limit to use when checking if type is valid
                              and indicating that some option was incorrect. */
};
typedef enum madcap_opt_type madcap_opt_type;

/* The format of a MADCAP message. */
struct madcap_msg_data {
    unsigned char version;
    unsigned char msgtype;
    uint16 addrfamily;
    uint32 xid;
    uint8 options[MAX_OPTIONS];
};

/* The madcap_msg object. */
struct madcap_msg_ {
    struct madcap_msg_data data;

    /* The length of the options field. */
    int options_len;
    uint32 sender_address;
    uint32 destination_address;
    /* Options included in message represented as binary flags. */
    uint16 opt_types;
    /* If some syntax error in message this variable indicates which option is
       incorrect. If impossible to state which option is incorrect it is set to
       MC_OPT_VALID (17). */
    madcap_opt_type incorrect;
    /* An array of pointers to options in message. */
    uint8 *options[MC_OPT_VALID];

    /* For further details on functions below check madcap_msg.c. */
    int (* create) (madcap_msg *object, madcap_msg_type msgtype,
                    uint16 addrfamily, uint32 xid,
                    uint8 *options, int optionslen);
    int (* parse) (madcap_msg *object, uint8 *data, int datalen,
                   net_peer *np);
    int (* send) (madcap_msg *object, net *n, net_socket *ns);
    int (* delete) (madcap_msg *object);
    int (* get_msg_type) (madcap_msg *object, madcap_msg_type *msgtype);
    int (* get_addrfamily) (madcap_msg *object, uint16 *addrfam);
    int (* get_xid) (madcap_msg *object, uint32 *xid);
    int (* get_sender) (madcap_msg *object, uint32 *sender);
    int (* get_destination) (madcap_msg *object, uint32 *destination);
    int (* get_multicasted) (madcap_msg *object, int16 *bool);
    int (* get_options_flag) (madcap_msg *object, uint16 *options);
    int (* get_option) (madcap_msg *object, madcap_opt_type opt,
                        uint16 *optlen, uint8 **optiondata);
    int (* get_error_option) (madcap_msg *object, madcap_opt_type *opt);
   
};

/**
 * The constructor of a madcap-msg object.
 *
 * @return A pointer to a madcap_msg object.
 */
madcap_msg *madcap_msg_new();

#endif
