/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains functions that is used by the message modules to
 * format the options in outgoing messages.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */


#include "madcap_msg_format_options.h"
extern int h_errno;
extern int gethostname(char *name, size_t len);


/*
 * Copies options to string format.
 * The function copies data from an struct with options to the data variable.
 * The data variable is then in its final form and is ready to be sent to a
 * client.
 *
 * @param opt_data An object with option in network byte order.
 * @param opt_len The length of the opt_data argument.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The function will reallocate the memory to the correct size.
 * @param data_len The length of the data.
 * @return 0 on success and -1 on failure.
 */
static int to_string(uint8 *opt_data,
                     int opt_len,
                     uint8 **data,
                     int *data_len) {

    *data     = (uint8 *) realloc(*data, sizeof(uint8) * (*data_len + opt_len));
    ASSERT(data != NULL, "Memory reallocation failed.");
    memcpy(*data + *data_len, opt_data, opt_len);
    *data_len = *data_len + opt_len;

    return 0;
}


/*
 * Gets server IP number.
 * The functions gets the IP number of the server.
 *
 * @param object The object that should be deleted.
 * @param server_id The IP number of the server in host byte order.
 * @return 0 on success and -1 on failure.
 */
static int pub_get_server_id(madcap_msg_format_options *object,
                             uint32 *server_id) {

    char server_name[256];
    struct hostent *server;
    int result = SUCCESS;
    
    /* Get server name and IP number. */
    result = gethostname(server_name, 256);
    if(result == SUCCESS) {
        server_name[255] = '\0';
        server  = gethostbyname(server_name);
        if(server == NULL) {
            result = FAILURE;
        }
    }

    if(result == SUCCESS) {
        /*
          Note that the server might have several IP numbers. The first one in
          the struct is returned.
         */
        if(server->h_addr_list[0] != NULL) {
            *server_id =
                (uint32)((struct in_addr *)server->h_addr_list[0])->s_addr;
            if(DEBUG_ENABLED('f') == SUCCESS) {
                printf("Server id = %s (%s)\n",
                       inet_ntoa(*(struct in_addr *) server_id), server_name);
            }
            *server_id = ntohl(*server_id);
        } else {
            result = FAILURE;
        }
    }

    return result;
}


/*
 * Check if an IP number is the server IP number.
 * The functions checks if an IP number belongs to the server.
 *
 * @param object The object that should be deleted.
 * @param server_id The IP number of the server in host byte order.
 * @return 0 on if server IP, -1 if not server IP.
 */
static int pub_is_server_id(madcap_msg_format_options *object,
                            uint32 server_id) {

    char server_name[256];
    struct hostent *server;
    uint32 serv_id;
    int result = SUCCESS;
    int i = 0;
    
    /* Get server name and IP number. */
    result = gethostname(server_name, 256);
    if(result == SUCCESS) {
        server_name[255] = '\0';
        server  = gethostbyname(server_name);
        if(server == NULL) {
            result = FAILURE;
        }
    }
    
    if(result == SUCCESS) {
        result = FAILURE;
        DEBUG('f',"Check match against server id:\n");
        while((server->h_addr_list[i] != NULL) && (result == FAILURE)) {
            serv_id =
                (uint32)((struct in_addr *)server->h_addr_list[i])->s_addr;
            if(serv_id == ntohl(server_id)) {
                if(DEBUG_ENABLED('f') == SUCCESS) {
                    printf("Match on server id = %s\n",
                           inet_ntoa(*(struct in_addr *) (&serv_id)));
                }                
                result = SUCCESS;
            } else {
                if(DEBUG_ENABLED('f') == SUCCESS) {
                    printf("Mismatch on server id = %s\n",
                           inet_ntoa(*(struct in_addr *) (&serv_id)));
                }
            }
            i++;
        }
    }

    return result;
}


/*
 * Deletes the object.
 * The function deletes the objectand returnes used resources to the system.
 *
 * @param object The object that should be deleted.
 * @return 0 on success and -1 on failure.
 */
static int pub_delete(madcap_msg_format_options *object) {

    free(object);

    return SUCCESS;
}


/*
 * Formats the END (0) option.
 *
 * @param object The object itself.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_end(madcap_msg_format_options *object,
                   uint8 **data,
                   int *data_len) {

    uint16 code;
    uint16 len;

    DEBUG('f',"Formating option: end.\n");

    code = htons(0);
    len  = htons(0);

    to_string((uint8 *) &code, 2, data, data_len);
    to_string((uint8 *) &len,  2, data, data_len);

    return SUCCESS;
}


/*
 * Formats the LEASE TIME (1) option.
 *
 * @param object The object itself.
 * @param lease_time The lease time to format.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_lease_time(madcap_msg_format_options *object,
                          uint32 lease_time,
                          uint8 **data,
                          int *data_len) {

    uint16 code;
    uint16 len;
    uint32 lease_tim;

    DEBUG('f',"Formating option: lease time.\n");

    code      = htons(1);
    len       = htons(4);
    lease_tim = htonl(lease_time);

    to_string((uint8 *) &code,      2, data, data_len);
    to_string((uint8 *) &len,       2, data, data_len);
    to_string((uint8 *) &lease_tim, 4, data, data_len);

    return SUCCESS;
}


/*
 * Formats the SERVER ID (2) option.
 *
 * @param object The object itself.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_server_id(madcap_msg_format_options *object,
                         uint8 **data,
                         int *data_len) {

    uint16 code;
    uint16 len;
    uint16 family;
    uint32 address;
    int result = SUCCESS;
    
    DEBUG('f',"Formating option: server id.\n");

    result = pub_get_server_id(NULL, &address);
    if(result == SUCCESS) {
        code    = htons(2);
        len     = htons(6);
        family  = htons(MC_ADDR_FAMILY);
        address = htonl(address);

        to_string((uint8 *) &code,    2, data, data_len);
        to_string((uint8 *) &len,     2, data, data_len);
        to_string((uint8 *) &family,  2, data, data_len);
        to_string((uint8 *) &address, 4, data, data_len);
    } else {
        mc_error->notify(mc_error, MC_ERR_OPTIONS,
                         "madcap_msg_format_options: Unknown server id.");
        DEBUG('f', "madcap_msg_format_options: Unknown server id.\n");
    }

    return result;
}


/*
 * Formats the LEASE ID (3) option.
 *
 * @param object The object itself.
 * @param client_lease_id A pointer to the lease identifier data to be included.
 * @param lease_id_len The length of the client_lease_id.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_lease_id(madcap_msg_format_options *object,
                        uint8 *client_lease_id,
                        int lease_id_len,
                        uint8 **data,
                        int *data_len) {

    uint16 code;
    uint16 len;
    uint8 lease_id[lease_id_len];
    int result = SUCCESS;

    DEBUG('f',"Formating option: lease id.\n");

    /* Check for incorrect arguments. */
    if(lease_id_len < 1) {
        mc_error->notify(mc_error, MC_ERR_OPTIONS,
                         "madcap_msg_format_options: Incorrect lease id.");
        DEBUG('f', "madcap_msg_format_options: Incorrect lease id.\n");
        result = FAILURE;
    } else {
        /* Format the option. */
        code   = htons(3);
        len    = htons(lease_id_len);
        memcpy(lease_id, client_lease_id, lease_id_len);
        
        to_string((uint8 *) &code,    2,            data, data_len);
        to_string((uint8 *) &len,     2,            data, data_len);
        to_string((uint8 *) lease_id, lease_id_len, data, data_len);
    }

    return result;
}


/*
 * Formats the MULTICAST SCOPE (4) option.
 *
 * @param object The object itself.
 * @param scope_id The scope id of the multicast scope.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_scope(madcap_msg_format_options *object,
                     uint32 scope_id,
                     uint8 **data,
                     int *data_len) {

    uint16 code;
    uint16 len;
    uint32 scope_idn;

    DEBUG('f',"Formating option: multicast scope.\n");

    code      = htons(4);
    len       = htons(4);
    scope_idn = htonl(scope_id);

    to_string((uint8 *) &code,      2, data, data_len);
    to_string((uint8 *) &len,       2, data, data_len);
    to_string((uint8 *) &scope_idn, 4, data, data_len);

    return SUCCESS;
}


/*
 * Formats the START TIME (6) option.
 *
 * @param object The object itself.
 * @param start_time The start time to format.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_start_time(madcap_msg_format_options *object,
                          uint32 start_time,
                          uint8 **data,
                          int *data_len) {

    uint16 code;
    uint16 len;
    uint32 start_tim;

    DEBUG('f',"Formating option: start time.\n");

    code      = htons(6);
    len       = htons(4);
    start_tim = htonl(start_time);

    to_string((uint8 *) &code,      2, data, data_len);
    to_string((uint8 *) &len,       2, data, data_len);
    to_string((uint8 *) &start_tim, 4, data, data_len);

    return SUCCESS;
}


/*
 * Formats the MULTICAST SCOPE LIST (9) option.
 *
 * @param object The object itself. 
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_scope_list(madcap_msg_format_options *object,
                          uint8 **data,
                          int *data_len) {

    /* TODO: Implement. */
    
    return FAILURE;
}


/*
 * Formats the LIST OF ADDRESS RANGES (10) option.
 *
 * @param object The object itself.
 * @param addr_range_list The address range lis to format.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_addr_range_list(madcap_msg_format_options *object,
                               madcap_plist *addr_range_list,
                               uint8 **data,
                               int *data_len) {
    
    uint16 code;
    uint16 len;
    uint32 addr;
    uint16 no;
    uint16 addr_range_list_len;
    addr_range *addr_rng;
    int result = SUCCESS;
    int i;

    DEBUG('f',"Formating option: list of address ranges.\n");

    /* Check that the madcap_plist is not empty. */
    if((addr_range_list == NULL) ||
       (addr_range_list->is_empty(addr_range_list) == 1) ||
       (addr_range_list->length(addr_range_list) < 1)) {
        result               = FAILURE;
    } else {
        /* Format the option. */
        addr_range_list_len  = addr_range_list->length(addr_range_list);
        
        code                 = htons(10);
        len                  = htons(addr_range_list_len * 6);

        to_string((uint8 *) &code, 2, data, data_len);
        to_string((uint8 *) &len,  2, data, data_len);

        /* Add address ranges one by one. */
        for(i = 0; i < addr_range_list_len; i++) {
            if(addr_range_list->is_empty(addr_range_list) == 1) {
                i            = addr_range_list_len;
                result       = FAILURE;
            } else {
                addr_rng     = addr_range_list->first(addr_range_list);
                addr         = addr_rng->first_addr(addr_rng);
                no           = addr_rng->last_addr(addr_rng) - addr + 1;
                addr_range_list->remove(addr_range_list, 0);
            
                addr         = htonl(addr);
                no           = htons(no);
        
                to_string((uint8 *) &addr, 4, data, data_len);
                to_string((uint8 *) &no,   2, data, data_len);
            }
        }
    }

    if(result == FAILURE) {
        mc_error->notify(mc_error, MC_ERR_OPTIONS,
                         "madcap_msg_format_options: No list of addr ranges.");
        DEBUG('f', "madcap_msg_format_options: No list of addr ranges.\n");
    }
    
    return result;
}


/*
 * Formats the CURRENT TIME (11) option.
 *
 * @param object The object itself.
 * @param current_time The current time to format.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_current_time(madcap_msg_format_options *object,
                            uint32 current_time,
                            uint8 **data,
                            int *data_len) {

    uint16 code;
    uint16 len;
    uint32 cur_time;

    DEBUG('f',"Formating option: current time.\n");

    code     = htons(11);
    len      = htons(4);
    cur_time = htonl(current_time);

    to_string((uint8 *) &code,     2, data, data_len);
    to_string((uint8 *) &len,      2, data, data_len);
    to_string((uint8 *) &cur_time, 4, data, data_len);

    return SUCCESS;
}


/*
 * Formats the FEATURE LIST (12) option.
 *
 * @param object The object itself. 
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_feature_list(madcap_msg_format_options *object,
                            uint8 **data,
                            int *data_len) {

    /* TODO: Implement. */
    
    return FAILURE;
}


/*
 * Formats the RETRY TIME (13) option.
 *
 * @param object The object itself. 
 * @param retry_time The time when a retry should be made.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */
static int pub_retry_time(madcap_msg_format_options *object,
                          uint32 retry_time,
                          uint8 **data,
                          int *data_len) {

    uint16 code;
    uint16 len;
    uint32 retry_tim;

    DEBUG('f',"Formating option: retry time.\n");

    code      = htons(13);
    len       = htons(4);
    retry_tim = htonl(retry_time);

    to_string((uint8 *) &code,      2, data, data_len);
    to_string((uint8 *) &len,       2, data, data_len);
    to_string((uint8 *) &retry_tim, 4, data, data_len);

    return SUCCESS;
}


/*
 * Formats the ERROR (16) option.
 *
 * @param object The object itself. 
 * @param err The error code indicating the cause for this option.
 * @param err_data A pointer to extra error data information that shall be
 * included in the formated option. The err_data must be in network byte order.
 * @param err_len The length of the err_data variable.
 * @param data A pointer to the location where a pointer to the resulting data
 * is placed. The memory will be reallocated to the correct size.
 * @param data_len A pointer to the location where the length of the resulting
 * data is placed.
 * @return 0 on success and -1 on failure.
 */

static int pub_error(madcap_msg_format_options *object, 
                     madcap_nak_error_code err_code,
                     uint8 *err_data,
                     int err_data_len,
                     uint8 **data,
                     int *data_len) {

    uint16 code;
    uint16 len;
    uint16 error_code;
    uint8 extra_data[err_data_len];
    int result = SUCCESS;
    
    DEBUG('f',"Formating option: error.\n");

    /* Check for incorrect arguments. */
    if(((err_code == MC_NAK_VALID_REQ_NOT_COMPLETED) && (err_data_len != 2)) |
       ((err_code == MC_NAK_INVALID_REQ) && (err_data_len != 2)) |
       ((err_code == MC_NAK_EXCESSIVE_CLK_SKEW) && (err_data_len != 4)) |
       ((err_code == MC_NAK_LEASE_ID_NOT_RECON) && (err_data_len != 0)) |
       ((err_code == MC_NAK_REQ_FEATURE_NOT_SUPPORT) && (err_data_len <0))) {
        mc_error->notify(mc_error, MC_ERR_OPTIONS,
                         "madcap_msg_format_options: Wrong length.");
        DEBUG('f',"madcap_msg_format_options: Wrong length.\n");
        result = FAILURE;
    } else if(err_code >= MC_NAK_VALID) {
        mc_error->notify(mc_error, MC_ERR_OPTIONS,
                         "madcap_msg_format_options: Invalid error code.");
        DEBUG('f',"madcap_msg_format_options: Invalid error code.\n");
        result = FAILURE;
    } else {
        /* Format the option. */
        code              = htons(16);
        len               = htons(2 + err_data_len);
        error_code        = htons(err_code);
        memcpy(extra_data, err_data, err_data_len);
        
        to_string((uint8 *) &code,       2,            data, data_len);
        to_string((uint8 *) &len,        2,            data, data_len);
        to_string((uint8 *) &error_code, 2,            data, data_len);
        to_string((uint8 *) extra_data,  err_data_len, data, data_len);
    }

    return result;
}


/**
 * Creates an object used to format options for outgoing messages.
 * 
 * @return The newly created object. The object should be deleted by the delete
 * function. A return value of NULL indicates a failure to create the new
 * message.
 */
madcap_msg_format_options *madcap_msg_format_options_new() {
    
    madcap_msg_format_options *object =
        (madcap_msg_format_options *) malloc(sizeof(madcap_msg_format_options));
    ASSERT(object != NULL, "Memory allocation failed.");

    object->end             = pub_end;
    object->lease_time      = pub_lease_time;
    object->server_id       = pub_server_id;
    object->lease_id        = pub_lease_id;
    object->scope           = pub_scope;
    object->start_time      = pub_start_time;
    object->scope_list      = pub_scope_list;
    object->addr_range_list = pub_addr_range_list;
    object->current_time    = pub_current_time;
    object->feature_list    = pub_feature_list;
    object->retry_time      = pub_retry_time;
    object->error           = pub_error;
    object->get_server_id   = pub_get_server_id;
    object->is_server_id    = pub_is_server_id;
    object->delete          = pub_delete;

    return object;
}

