/*
 * This is the maasd program, a part of an implementation of the MALLOC
 * architecture.
 *
 * This file contains functions that is used when the server receives a
 * REQUEST message.
 *
 * Copyright (C) 2001 Hkan Bystrm,Maria Isberg, Pia Lindstm,
 * Fredrik Pettersson, Linus Svensson and ke stmark.
 *
 * The project is supported by Telia Research AB, Lule, Sweden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 1307 USA
 *
 * The authors of this program can be reached on malloc@deimos.campus.luth.se,
 * or via the project homepage at http://deimos.campus.luth.se/malloc/.
 * The Telia representative is Thomas Eriksson (Thomas.A.Eriksson@telia.se)
 */


#include "madcap_msg_request.h"
static int pub_receive(madcap_msg_request *object);
static int pub_delete(madcap_msg_request *object);
extern madcap_error *mc_error;

typedef enum request_type_ request_type;

enum request_type_ {
    UNKNOWN,           /* Should not be used. */
    UNICAST_ALLOC,     /* Unicast allocation (ACK/NAK as response). */
    UNICAST_ERROR,     /* Invalid unicast allocation (NAK as response). */
    MULTICAST_CONFIRM, /* Confirmation to this server (ACK/NAK as response). */
    MULTICAST_RELEASE, /* Confirmation to another server (no response). */
    MULTICAST_ERROR    /* Invalid multicast message (no response) */
};

/**
 * Creates a new REQUEST message.
 * 
 * @param msg An incoming madcap_msg message with madcap_msg_type = MC_REQUEST.
 * @param cfg The configuation settings for the madcap server.
 * @param addr The address record of the madcap server.
 * @param n A network abstraction object.
 * @param ns An abstract network connection.
 * @param mc The madcap object.
 * @return The newly created message. The message should be deleted using the
 * delete function. A return value of NULL indicates a failure to create the
 * new message.
 */
madcap_msg_request *madcap_msg_request_new(madcap_msg *msg,
                                           madcap_cfg *cfg,
                                           addr_rec *addr,
                                           net *n,
                                           net_socket *ns,
                                           madcap *mc) {

    madcap_msg_request *object;

    DEBUG('m',"Creating REQUEST message.\n");
    
    object = (madcap_msg_request *) malloc(sizeof(madcap_msg_request));
    ASSERT(object != NULL, "Memory allocation failed.");
    
    object->msg     = msg;
    msg->get_options_flag(msg, &(object->opt_types_incl));
    object->cfg     = cfg;
    object->addr    = addr;
    object->n       = n;
    object->ns      = ns;
    object->mc      = mc;
    msg->get_xid(msg, &(object->xid));
    object->receive = pub_receive;
    object->delete  = pub_delete;

    return object;    
}


/*
 * Receives a REQUEST message.
 * The function takes the appropriate actions when a REQUEST message is
 * received. The function will send either an ACK or an NAK to the client as
 * response.
 *
 * @param object The message (object) that is received.
 * @return 0 on success and -1 on failure.
 */
static int pub_receive(madcap_msg_request *object) {

    madcap_msg_ack *ack;
    madcap_msg_nak *nak;
    madcap_msg_format_options *format = madcap_msg_format_options_new();
    int send_ack                   = TRUE;
    int unicast                    = TRUE;
    request_type msg_type          = UNKNOWN;
    int result                     = SUCCESS;
    int ok;                         /* Temp variable. */
    int check;                      /* Temp variable. */
    int i;                          /* Temp variable. */
    uint8 *opt_data;                /* Option data from incoming messages. */
    uint16 opt_data_len;            /* Length of opt_data. */

    uint32 lease_time_req          = 0;    /* Used with option 1. */
    uint32 lease_time_ack          = 0;    /* Used with option 1. */
    uint32 server_id               = 0;    /* Used with option 2. */
    uint16 addr_family             = 0;    /* Used with option 2. */
    uint8 *lease_id                = NULL; /* Used with option 3. */
    uint16 lease_id_len            = 0;    /* Used with option 3. */
    uint32 scope_id                = 0;    /* Used with option 4. */
    scope *scopes                  = NULL; /* Used with option 4. */
    int scope_count                = 0;    /* Used with option 4. */
    uint32 start_time_req          = 0;    /* Used with option 6. */
    uint32 start_time_ack          = 0;    /* Used with option 6. */
    uint16 no_addr                 = 0;    /* Used with option 7. */
    uint16 min_no_addr             = 0;    /* Used with option 7. */
    uint32 first_addr              = 0;    /* Used with option 10. */
    uint32 last_addr               = 0;    /* Used with option 10. */
    uint16 addr_no                 = 0;    /* Used with option 10. */
    addr_range *addr_rng           = NULL; /* Used with option 10. */
    madcap_plist *addr_ranges_req  = NULL; /* Used with option 10. */
    madcap_plist *addr_ranges_ack  = NULL; /* Used with option 10. */
    uint32 cur_time_client         = 0;    /* Used with option 11. */
    uint32 cur_time_server         = 0;    /* Used with option 11. */
    uint32 clk_skew_allow          = 0;    /* Used with option 11. */
    uint32 min_lease_time          = 0;    /* Used with option 14. */
    uint32 max_start_time          = 0;    /* Used with option 15. */
    madcap_nak_error_code err_code = 0;    /* Used with option 16. */
    uint16 err_option              = 0;    /* Used with option 16. */
    int err_option_count           = 0;    /* Used with option 16. */
    uint32 err_time                = 0;    /* Used with option 16. */
    int err_data_len               = 2;    /* Used with option 16. */
    madcap_error_type err_num      = 0;    /* Used with option 16. */

    DEBUG('m',"Receiving REQUEST message.\n");

    /* Check if message is an unicast or a multicast message. */
    if(ntohl(object->n->sender->dest_addr) < ntohl(inet_addr("224.0.0.0"))) {
        unicast = TRUE;
    } else {
        unicast = FALSE;    
    }

    /* Check if optional option LEASE TIME (1) is present and valid. */
    ok = (object->opt_types_incl & 0x0001) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_LEASE_TIME,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len == 4) && (opt_data != NULL)) {
            memcpy(&lease_time_req, opt_data, opt_data_len);
            lease_time_req = ntohl(lease_time_req);
        } else {
            err_code       = MC_NAK_INVALID_REQ;
            err_option     = 1;
            err_option_count++;
            send_ack       = FALSE;
            DEBUG('m',"Lease time present but not valid.\n");
        }
    } else {
        lease_time_req     = 0;
    }

    if(unicast == TRUE) {
        /*
          Check if option SERVER IDENTIFIER (2) is present and valid. The
          option is optional for unicast messages.
        */
        msg_type                   = UNICAST_ALLOC;
        ok = (object->opt_types_incl & 0x0002) != 0;
        if(ok) {
            check = object->msg->get_option(object->msg, MC_SERVER_ID,
                                            &opt_data_len, &opt_data);
            if((check == SUCCESS) && (opt_data_len == 6) && (opt_data != NULL)){
                memcpy(&addr_family, opt_data, 2);
                addr_family = ntohs(addr_family);
                memcpy(&server_id, opt_data + 2, 4);
                server_id = ntohl(server_id);
                check = format->is_server_id(format, server_id);
                if(addr_family != MC_ADDR_FAMILY) {
                    msg_type       = UNICAST_ERROR;
                    err_code       = MC_NAK_INVALID_REQ;
                    err_option     = 2;
                    err_option_count++;
                    send_ack       = FALSE;
                    DEBUG('m',"Wrong address family.\n");
                } else if(check != SUCCESS) {
                    msg_type   = UNICAST_ERROR;
                    err_code   = MC_NAK_INVALID_REQ;
                    err_option = 2;
                    err_option_count++;
                    send_ack   = FALSE;
                    DEBUG('m',"Provided server id is not this server.\n");
                }
            } else {
                msg_type           = UNICAST_ERROR;
                err_code           = MC_NAK_INVALID_REQ;
                err_option         = 2;
                err_option_count++;
                send_ack           = FALSE;
                DEBUG('m',"Server id present but not valid.\n");
            }
        }
    } else {
        /*
          Check if option SERVER IDENTIFIER (2) is present and valid. The
          option is required for multicast messages.
        */
        msg_type                   = MULTICAST_ERROR;
        ok = (object->opt_types_incl & 0x0002) != 0;
        if(ok) {
            check = object->msg->get_option(object->msg, MC_SERVER_ID,
                                            &opt_data_len, &opt_data);
            if((check == SUCCESS) && (opt_data_len == 4) && (opt_data != NULL)){
                memcpy(&addr_family, opt_data, 2);
                addr_family = ntohs(addr_family);
                memcpy(&server_id, opt_data + 2, 4);
                server_id = ntohl(server_id);
                check = format->is_server_id(format, server_id);
                if(addr_family != MC_ADDR_FAMILY) {
                    err_code       = MC_NAK_INVALID_REQ;
                    err_option     = 2;
                    err_option_count++;
                    send_ack       = FALSE;
                    DEBUG('m',"Wrong address family.\n");
                } else if(check == SUCCESS) {
                    msg_type       = MULTICAST_CONFIRM;
                    DEBUG('m',"Message is a confirmation to this server.\n");
                } else {
                    msg_type       = MULTICAST_RELEASE;
                    DEBUG('m',"Message is a confirmation to another server.\n");
                }
            }
        }
        if(msg_type == MULTICAST_ERROR) {
            DEBUG('m',
                  "Invalid multicast message, no response sent to client.\n");
        }
    }
        
    /* Check that required option LEASE ID (3) is present and valid. */
    ok = (object->opt_types_incl & 0x0004) != 0;
    if(ok) { 
        check = object->msg->get_option(object->msg, MC_LEASE_ID,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len > 0) && (opt_data != NULL)) {
            lease_id     = opt_data;
            lease_id_len = opt_data_len;
        } else {
            ok           = FALSE;
            DEBUG('m',"Lease ID present but not valid.\n");
        }
    }
    if(!ok) {
        err_code         = MC_NAK_INVALID_REQ;
        err_option       = 3;
        err_option_count++;
        send_ack         = FALSE;
        DEBUG('m',"Lease ID not present.\n");
    }
    
    /* Check that required option MULTICAST SCOPE (4) is present and valid. */
    ok = (object->opt_types_incl & 0x0008) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_SCOPE,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len == 4) && (opt_data != NULL)) {
            memcpy(&scope_id, opt_data, opt_data_len);
            scope_id = ntohl(scope_id);
            object->cfg->get_number_of_scopes(object->cfg, &scope_count);
            object->cfg->get_scopes(object->cfg, &scopes);
            check         = FAILURE;
            for(i = 0;i < scope_count;i++) {
                if(memcmp(&scope_id, &(scopes[i]), 4) == 0) {
                    check = SUCCESS;
                    i     = scope_count;
                }
            }
            if(check != SUCCESS) {
                ok        = FALSE;
                DEBUG('m',"Multicast scope present but not valid.\n");
            }
        } else {
            ok            = FALSE;
            DEBUG('m',"Multicast scope present but not valid.\n");
        }
    } else {
        DEBUG('m',"Multicast scope not present.\n");
    }
    if(!ok) {
        err_code          = MC_NAK_INVALID_REQ;
        err_option        = 4;
        err_option_count++;
        send_ack          = FALSE;
    }

    /* Check if optional option START TIME (6) is present and valid. */
    ok = (object->opt_types_incl & 0x0020) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_START_TIME,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len == 4) && (opt_data != NULL)) {
            memcpy(&start_time_req, opt_data, opt_data_len);
            start_time_req = ntohl(start_time_req);
        } else {
            err_code       = MC_NAK_INVALID_REQ;
            err_option     = 6;
            err_option_count++;
            send_ack       = FALSE;
            DEBUG('m',"Start time present but not valid.\n");
        }
    } else {
        start_time_req     = 0;
    }
    
    /*
     * Check if optional option NUMBER OF ADDRESSES REQUESTED (7) is present
     * and valid.
     */
    ok = (object->opt_types_incl & 0x0040) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_NO_ADDR_REQ,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len == 4) && (opt_data != NULL)) {
            memcpy(&no_addr, opt_data + 2, 2);
            memcpy(&min_no_addr, opt_data, 2);
            no_addr        = ntohs(no_addr);
            min_no_addr    = ntohs(min_no_addr);
            if((no_addr < 1) || (min_no_addr < 1) || (no_addr < min_no_addr)) {
                check      = FAILURE;
            }
        }
        if(check != SUCCESS) {
            err_code       = MC_NAK_INVALID_REQ;
            err_option     = 7;
            err_option_count++;
            send_ack       = FALSE;
            DEBUG('m',
                  "Number of addresses requested present but not valid.\n");
        }
    } else {
        no_addr            = 1;
        min_no_addr        = 1;
    }

    /*
     * Check if optional option LIST OF ADDRESS RANGES (10) is present and
     * valid.
     */
    ok = (object->opt_types_incl & 0x0200) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_ADDR_RANGE_LIST,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) &&
           ((opt_data_len % 6) == 0) &&
           (opt_data != NULL)) {
            addr_ranges_req = (madcap_plist *) malloc(sizeof(madcap_plist));
            addr_ranges_req = madcap_plist_new();
            /* Add address ranges one by one. */
            for(i = 0; i < opt_data_len / 6; i++) {
                memcpy(&first_addr, opt_data + i * 6, 4);
                memcpy(&addr_no, opt_data + i * 6 + 4, 2);
                addr_no     = ntohs(addr_no);
                last_addr   = htonl(ntohl(first_addr) + addr_no);
                addr_rng    = addr_range_new(first_addr, last_addr, 0);
                addr_ranges_req->insert(addr_ranges_req,
                                        ntohl(first_addr),
                                        addr_rng);
            }
        } else {
            err_code        = MC_NAK_INVALID_REQ;
            err_option      = 10;
            err_option_count++;
            send_ack        = FALSE;
            DEBUG('m',"List of address ranges present but not valid.\n");
        }
    }

    /* Check if optional option CURRENT TIME (11) is present and valid. */
    ok = (object->opt_types_incl & 0x0400) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_CURRENT_TIME,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len == 4) && (opt_data != NULL)) {
            memcpy(&cur_time_client, opt_data, opt_data_len);
            cur_time_client = ntohl(cur_time_client);
            /* Check clock skew. */
            time(&cur_time_server);
            object->cfg->get_clock_skew_allowance(object->cfg,
                                                  object->cfg->name,
                                                  &clk_skew_allow);
            if(labs(cur_time_client - cur_time_server) > clk_skew_allow) {
                err_code    = MC_NAK_EXCESSIVE_CLK_SKEW;
                err_time    = cur_time_server;
                err_option  = 11;
                err_option_count++;
                send_ack    = FALSE;
                DEBUG('m',"Clock skew detected.\n");
            }

        } else {
            err_code        = MC_NAK_INVALID_REQ;
            err_option      = 11;
            err_option_count++;
            send_ack        = FALSE;
            DEBUG('m',"Current time present but not valid.\n");
        }
    }
    
    /* Check if optional option FEATURE LIST (12) is present and valid. */
    /* TODO: */

    /* Check if optional option MINIMUM LEASE TIME (14) is present and valid. */
    ok = (object->opt_types_incl & 0x2000) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_MIN_LEASE_TIME,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len == 4) && (opt_data != NULL)) {
            memcpy(&min_lease_time, opt_data, opt_data_len);
            min_lease_time = ntohl(min_lease_time);
        } else {
            err_code       = MC_NAK_INVALID_REQ;
            err_option     = 14;
            err_option_count++;
            send_ack       = FALSE;
            DEBUG('m',"Min lease time present but not valid.\n");
        }
    } else {
        min_lease_time = 0;
    }
    
    /* Check if optional option MAXIMUM START TIME (15) is present and valid. */
    ok = (object->opt_types_incl & 0x4000) != 0;
    if(ok) {
        check = object->msg->get_option(object->msg, MC_MAX_START_TIME,
                                        &opt_data_len, &opt_data);
        if((check == SUCCESS) && (opt_data_len == 4) && (opt_data != NULL)) {
            memcpy(&max_start_time, opt_data, opt_data_len);
            max_start_time = ntohl(max_start_time);
        } else {
            err_code       = MC_NAK_INVALID_REQ;
            err_option     = 15;
            err_option_count++;
            send_ack       = FALSE;
            DEBUG('m',"Max start time present but not valid.\n");
        }
    } else {
        max_start_time = 0;
    }
    
    /*
     * Check if optional option CURRENT TIME (11) is present if required by
     * optional option START TIME (6) and/or MAXIMUM START TIME (15).
     */
    if(((object->opt_types_incl & 0x4020) != 0) &&
       ((object->opt_types_incl & 0x0400) == 0)) {
        err_code       = MC_NAK_INVALID_REQ;
        err_option     = 11;
        err_option_count++;
        send_ack       = FALSE;
        DEBUG('m',"Current time not present.\n");
    } else {
        /* Check if optional option START TIME (6) is in the past. */
        if(((object->opt_types_incl & 0x0020) != 0) &&
           (start_time_req + clk_skew_allow < cur_time_server)) {
            err_code   = MC_NAK_INVALID_REQ;
            err_option = 6;
            err_option_count++;
            send_ack   = FALSE;
            DEBUG('m',"Start time is in the past.\n");
        }
        
        /* Check if optional option MAXIMUM START TIME (15) is in the past. */
        if(((object->opt_types_incl & 0x4000) != 0) &&
           (max_start_time + clk_skew_allow < cur_time_server)) {
            err_code   = MC_NAK_INVALID_REQ;
            err_option = 15;
            err_option_count++;
            send_ack   = FALSE;
            DEBUG('m',"Maximum start time is in the past.\n");
        }
    }
    
    /*
     * Check that no forbiden options are present.
     * madcap_opt_type 5, 8, 9, 13 and 16 are not allowed.
     */
    if((object->opt_types_incl & 0x9190) != 0) {
        err_code       = MC_NAK_INVALID_REQ;
        send_ack       = FALSE;
        if((object->opt_types_incl & 0x0010) != 0) {
            err_option = 5;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0080) != 0) {
            err_option = 8;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x0100) != 0) {
            err_option = 9;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x1000) != 0) {
            err_option = 13;
            err_option_count++;
        }
        if((object->opt_types_incl & 0x8000) != 0) {
            err_option = 16;
            err_option_count++;
        }
        DEBUG('m',"Forbiden option present.\n");
    }

    /* Process the request. */
    if(msg_type == MULTICAST_RELEASE) {
        object->addr->release_prealloc(object->addr, lease_id, lease_id_len);
    } else if(msg_type == MULTICAST_CONFIRM) {
        /*
          TODO: Implement a way to confirm a preallocation. This should be done
          with the confirm_prealloc() function in the addr_rec module. An
          alternative solution might be to use the alloc_addr() function
          instead.
          
          The response from this action is an ACK or NAK, so the code should be
          combined with the code below that handles UNICAST_ALLOC /
          UNICAST_ERROR messages.
        */
    } else if((msg_type == UNICAST_ALLOC) || (msg_type == UNICAST_ERROR)) {
        if(send_ack == TRUE) {
            addr_ranges_ack = (madcap_plist *) malloc(sizeof(madcap_plist));
            addr_ranges_ack = madcap_plist_new();
            
            ok =  object->addr->alloc_addr(object->addr,
                                           MADCAP_ALLOC_ALLOC,
                                           lease_id,
                                           lease_id_len,
                                           scope_id,
                                           lease_time_req,
                                           min_lease_time,
                                           start_time_req,
                                           max_start_time,
                                           no_addr,
                                           min_no_addr,
                                           addr_ranges_req,
                                           &lease_time_ack,
                                           &start_time_ack,
                                           addr_ranges_ack);
            if(ok == FAILURE) {
                send_ack           = FALSE;
                mc_error->madcap_errnum_collect(mc_error, &err_num);
                switch(err_num) {
                case MC_ERR_OUTOFADDR:
                    err_code       = MC_NAK_VALID_REQ_NOT_COMPLETED;
                    err_option     = 7;
                    err_option_count++;
                    break;
                case MC_ERR_ADDRLTIME:
                    err_code       = MC_NAK_VALID_REQ_NOT_COMPLETED;
                    if(lease_time_req > 0) {
                        err_option = 1;
                        err_option_count++;
                    }
                    if(start_time_req > 0) {
                        err_option = 6;
                        err_option_count++;
                    }
                    if(min_lease_time > 0) {
                        err_option = 14;
                        err_option_count++;
                    }
                    if(max_start_time > 0) {
                        err_option = 15;
                        err_option_count++;
                    }
                    break;
                default:
                    err_code       = MC_NAK_INVALID_REQ;
                    err_option     = 0xffff;
                    err_option_count++;
                }
            }
        }
    
        /* Create ACK or NAK message. */
        object->data     = NULL;
        object->data_len = 0;
        ok               = SUCCESS;
        
        /* Add the LEASE TIME (1) option (ACK). */
        if(send_ack == TRUE) {
            ok = ok + format->lease_time(format,
                                         lease_time_ack,
                                         &(object->data),
                                         &(object->data_len));
        }
        
        /* Add the SERVER IDENTIFIER (2) option (ACK and NAK). */
        ok = ok + format->server_id(format,
                                    &(object->data),
                                    &(object->data_len));
        
        /* Add the LEASE IDENTIFIER (3) option (ACK and NAK). */
        ok = ok + format->lease_id(format,
                                   lease_id,
                                   lease_id_len,
                                   &(object->data),
                                   &(object->data_len));
        
        /* Add the MULTICAST SCOPE (4) option (ACK). */
        if(send_ack == TRUE) {
            ok = ok + format->scope(format,
                                    scope_id,
                                    &(object->data),
                                    &(object->data_len));
        }
        
        /* Add the START TIME (6) option (ACK). */
        if((send_ack == TRUE) && (start_time_ack > 0)) {
            ok = ok + format->start_time(format,
                                         start_time_ack,
                                         &(object->data),
                                         &(object->data_len));
            if(cur_time_server == 0) {
                time(&cur_time_server);
            }
        }
        
        /* Add the LIST OF ADDRESS RANGES (10) option (ACK). */
        if(send_ack == TRUE) {
            ok = ok + format->addr_range_list(format,
                                              addr_ranges_ack,
                                              &(object->data),
                                              &(object->data_len));
        }
        
        /* Add the CURRENT TIME (11) option (ACK). */
        if((send_ack == TRUE) && (cur_time_server > 0)) {
            ok = ok + format->current_time(format,
                                           cur_time_server,
                                           &(object->data),
                                           &(object->data_len));
        }
        
        /* Add the FEATURE LIST (12) option (ACK and NAK). */
        /* TODO: */

        /* Add the RETRY TIME (13) option (ACK). */
        /* TODO: */
        
        /* Add the ERROR (16) option (NAK). */
        if(send_ack == FALSE) {
            if((err_code == MC_NAK_INVALID_REQ) ||
               (err_code == MC_NAK_VALID_REQ_NOT_COMPLETED)) {
                if(err_option_count != 1) {
                    /* Problem was not with a specific option. */
                    err_option = 0xffff;
                }
                err_option = htons(err_option);
                ok = ok + format->error(format,
                                        err_code,
                                        (uint8 *) &err_option,
                                        err_data_len,
                                        &(object->data),
                                        &(object->data_len));
            } else if(err_code == MC_NAK_EXCESSIVE_CLK_SKEW) {
                err_time = htonl(err_time);
                ok = ok + format->error(format,
                                        err_code,
                                        (uint8 *) &err_time,
                                        err_data_len,
                                        &(object->data),
                                        &(object->data_len));
            } else {
                ok = ok + format->error(format,
                                        err_code,
                                        (uint8 *) NULL,
                                        err_data_len,
                                        &(object->data),
                                        &(object->data_len));
            }
        }
        
        /* Add the END (0) option (ACK and NAK). */
        ok = ok + format->end(format,
                              &(object->data),
                              &(object->data_len));
        
        /* Send message if options were added correctly. */
        if((ok == SUCCESS) && (send_ack == TRUE)) {
            /* Send ACK. */
            ack = madcap_msg_ack_new(object->n,
                                     object->ns,
                                     object->xid,
                                     object->data,
                                     object->data_len);
            if(ack != NULL) {
                ok = ack->send(ack);
                if(ok == SUCCESS) {
                    /* Save message so it can be resent later if necessary. */
                    object->mc->save_outgoing_msg(object->mc,
                                                  object->xid,
                                                  MC_REQUEST,
                                                  MC_ACK,
                                                  lease_id,
                                                  lease_id_len,
                                                  ack);
                }
            } else {
                ok = FAILURE;
            }
            result = ok;
        } else if((ok == SUCCESS) && (send_ack == FALSE)) {
            /* Send NAK. */
            nak = madcap_msg_nak_new(object->n,
                                     object->ns,
                                     object->xid,
                                     object->data,
                                     object->data_len);
            if(nak != NULL) {
                ok = nak->send(nak);
                if(ok == SUCCESS) {
                    /* Save message so it can be resent later if necessary. */
                    object->mc->save_outgoing_msg(object->mc,
                                                  object->xid,
                                                  MC_REQUEST,
                                                  MC_NAK,
                                                  lease_id,
                                                  lease_id_len,
                                                  nak);
                }
            } else {
                ok = FAILURE;
            }
            result = ok;
        }
        
        /* Create error message if something failed. */
        if(ok == FAILURE) {
            if(send_ack == TRUE) {
                mc_error->notify(mc_error, MC_ERR_CREATE_ACK,
                                 "madcap_msg_request: Failed to create & send");
                result = FAILURE;
            } else {
                mc_error->notify(mc_error, MC_ERR_CREATE_NAK,
                                 "madcap_msg_request: Failed to create & send");
                result = FAILURE;
            }
        }
    }
    
    format->delete(format);
    free(addr_ranges_req);
    free(addr_ranges_ack);

    return result;
}


/*
 * Deletes a REQUEST message.
 * The function deletes a REQUEST message and returnes used resources to the
 * system.
 *
 * @param object The message (object) that should be deleted.
 * @return 0 on success and -1 on failure.
 */
static int pub_delete(madcap_msg_request *object) {

    free(object->data);
    free(object);

    return SUCCESS;
}


