#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <sys/time.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/uio.h>

#include "net.h"
#include "globals.h"

#define UMAX(a, b)      ((a) > (b) ? (a) : (b))

/**
 * Implements an abstraction of the network and methods
 * to receive and send data over the network. See net.h
 * for details.
 */

static int main_loop(net *n);
static void receive(net *n,net_socket *s, net_callback, void *obj);
static void unreceive(net *n, net_socket *ns);
static int send_msg(net *n, net_socket *s, char *data, int length);
extern int gethostname(char *name, size_t len);

/**
 * The registration struct implements the object that
 * holds all data for a registered socket i.e. the net_socket,
 * the callback function and the first argument in the
 * callback function. This structure is used in the list.
 */

typedef struct registration_ registration;

struct registration_{
    
    net_socket *net_sckt;
    net_callback callback;
    void *obj; 

};

/*------------------------------------------------------------------*/

/**
 * static int main_loop(net *);
 *
 *   The main_loop function implements a infinite loop listening
 *   for network data to arrive to some of the registered sockets
 *   or for a time event to occur.
 */
 
static int main_loop(net *n){
  
    struct msghdr mh;
    struct cmsghdr *cmsg;
    struct iovec miov;
    uint8 conbuf[1000];
    struct in_pktinfo *nfo;
    uint8 *buffer = NULL;
    int i, retval;
    list *regs;  
    int len, buffsize;
    char buff[1000];
    registration *sc;
    net_socket *tmp_ns;
    int max_desc;
    struct sockaddr_in from;
    int fromlen;
    char error_string[100];
    int size = 0;
    
    len = 1000;
    /*bzero(abuf, abuf_len);*/
    buffer = (uint8 *)malloc(1024 * sizeof(uint8));
    ASSERT(buffer != NULL, "Memory allocation failed.");
    
    regs = n->registrations;

    /* Initiate the infinite loop */
    for(; ;){
        fd_set in_set;     
        struct timeval time;
        struct timeval *time_ptr;
        
        /* are there any time events to be scheduled*/
        if( n->timer->is_empty(n->timer) ){
            time_ptr = NULL;
        }else{
            /* get the time when the next event occurs */
            n->timer->get_next_time(n->timer, &time); 
            time_ptr = &time;
            
            /* schedule each time event that already should
               have been scheduled. */	  	  
            while( !n->timer->is_empty(n->timer) &&
                   (time.tv_sec <= 0) && (time.tv_usec <= 0)){
                
                /* time event has already occured,
                   call the associated function */ 
                if( n->timer->call_next(n->timer) == -1){	    
                    perror("Error in timeout, no time_event");
                    exit(-1);	  
                }
                
                /* get the next timeout */
                n->timer->get_next_time(n->timer, &time);
            }
            /* again, check if there is no more time event */
            if(n->timer->is_empty(n->timer)) {
                time_ptr = NULL;
            }
            
        }      
        
                
        FD_ZERO(&in_set);
        max_desc = 0;
        

        /* put each registered socket in the fd_set */
        for( i = 0; i < regs->length(regs); i++){
            sc = (registration *)(regs->elem_at(regs,i));
            tmp_ns = (net_socket *)(sc->net_sckt);	          
            FD_SET(tmp_ns->socket, &in_set);                 
            max_desc = UMAX(max_desc, tmp_ns->socket);
        }

        /* wait for arriving network data or until some time event occurs */
        if((retval = select(max_desc+1, &in_set, NULL, NULL, time_ptr)) == 0){
            /* timeout occured*/
            if( n->timer->call_next(n->timer) == -1){
                perror("Error in timeout, no time_event");
                exit(-1);
            }
            
        }else{
            
            if (retval > 0){
                /* some of the registered sockets has received data */
                
                /* check which sockets that received the data */
                for(i = 0; i < regs->length(regs) ; i++){
                    
                    sc = (registration *)(regs->elem_at(regs,i));
                    
                    if( FD_ISSET(((sc->net_sckt)->socket), &in_set)){

                        /* Initiating struct msghdr. */
                        mh.msg_name    = NULL;
                        mh.msg_namelen = 0;
                        mh.msg_iov = &miov;
                        mh.msg_iovlen = 1;
                        mh.msg_control = conbuf;
                        mh.msg_controllen = 1000;
                        mh.msg_flags = 0;
                        miov.iov_base = buffer;
                        miov.iov_len = 1024;
                        
                        /* Peek at the arrived data from the socket to get
                           the destination address used by the sender. */
                        size = recvmsg(sc->net_sckt->socket, &mh, MSG_PEEK);
                        if(size > -1){
                            for (cmsg = CMSG_FIRSTHDR(&mh);
                                 cmsg != NULL;
                                 cmsg = CMSG_NXTHDR(&mh,cmsg)){
                                if (cmsg->cmsg_level == SOL_IP &&
                                    cmsg->cmsg_type == IP_PKTINFO) {
                                    struct in_addr d_addr;
                                    nfo = (struct in_pktinfo *)CMSG_DATA(cmsg);
                                    n->sender->dest_addr = nfo->ipi_addr.s_addr;
                                    d_addr.s_addr = n->sender->dest_addr;
                                    DEBUG('n', "Destination address of incoming message: %s\n", inet_ntoa(d_addr));
                                }
                            }
                        } else {
                            perror("net.c : Error in recvmsg");
                        }
                        
                        /* Read the arrived data from the socket. */
                        fromlen = sizeof(struct sockaddr_in);                  
                        buffsize = recvfrom(sc->net_sckt->socket, buff, len, 0,
                                            (struct sockaddr *)&from, &fromlen);
                        /* Prints all bytes received in verbose mode. */
                        if(DEBUG_ENABLED('s') == 0) {
                            printf("Receiving %d bytes from %s: ",
                                   buffsize,
                                   inet_ntoa(from.sin_addr));
                            for(i = 0; i < buffsize; i++) {
                                printf("%X,", (uint8) buff[i]);
                            }
                            printf("\n");
                        }
                        
                        if(  buffsize == -1){
                            perror("Error occured when revceiving msg from socket (recvfrom)");
                            exit(-1);
                            
                        } else {
                            
                            /* Get sender info. */       
                            strcpy(n->sender->ip_addr, inet_ntoa(from.sin_addr));
                            n->sender->peer_port = ntohs(from.sin_port);
                            n->sender->peer_addr = from.sin_addr.s_addr;
                            
                            /* Call the socket's callback function. */
                            retval = sc->callback(sc->obj, buff, buffsize,
                                                  n->sender);
                            
                            /* If something went wrong, print the cause. */
                            if (retval == -1){
                                DEBUG('e',"net: Error while processing sender message: ");
                                mc_error->madcap_errstring_collect(mc_error,
                                                                   error_string);
                                sprintf(error_string,"%s%s",error_string,"\n");
                                DEBUG('e',error_string);
                            }
                        }
                    }
                }
                
            }else{
                /* select error */
                perror("main_loop: Select failed!");
                exit(-1);
            }
        }            
    }      
}

/*------------------------------------------------------------------*/

/**
 * static void receive(net *n,net_socket *ns, net_callback nc, void *obj);
 *
 *   Register the callback function nc to be called whenever datagrams
 *   are received on the net_socket ns. obj will be passed as the first
 *   argument to nc when called.
 *
 */

static void receive(net *n,net_socket *ns, net_callback nc,
                    void *obj){
    
    registration *sc = (registration *)malloc(sizeof(registration));
    sc->net_sckt = ns;
    sc->callback = nc;
    sc->obj = obj;
    
    /**
     * insert the created registration
     * object into the registrations list
     */  
    n->registrations->insert(n->registrations, sc);
    n->nreg++;
    
}

/*------------------------------------------------------------------*/

/**
 * static void unreceive(net *n, net_socket *ns);
 *
 *   Unregisters the callback function identified by id. This identifier
 *   is obtained when invoking the receive method.
 */
static void unreceive(net *n, net_socket *ns){
    
    int i;
    registration *tmp;
    list *regs = n->registrations;
    
    /* step through the registrations list
     * to find the registration object
     * for the given net_socket and remove it
     * from the list.
     */
    for(i = 0 ; i < regs->length(regs); i++){
        
        tmp = (registration*)(regs->elem_at(regs,i));
        
        if( (net_socket *)(tmp->net_sckt) == ns ){
            free(tmp->obj);      
            free(tmp);
            regs->remove(regs, i);   
            break;
        }
    }
    
    /* decrease the number of registered net_sockets */ 
    n->nreg--;
    /* free memory for the net_socket object */ 
    ns->delete(ns);    
    
}

/*------------------------------------------------------------------*/

/**
 * static int send_msg(net *n, net_socket *s, char *data, int length);
 *
 *   Sends a datagram. The net_socket ns must be set up prior to the call to
 *   send. See the above example for details.
 */

static int send_msg(net *n, net_socket *s, char *data,
                    int length){

    int i; /* FIXME: REMOVE THIS CODE. */
    struct sockaddr *sa_tmp = (struct sockaddr *)&s->sa_dest;
  
    /* Prints all sent data. */
    if(DEBUG_ENABLED('s') == 0) {
        printf("Sending %d bytes: ", length);
        for(i = 0; i < length; i++) {
            printf("%X,", (uint8) data[i]);
        }
        printf("\n");
    }

    if( sendto(s->socket, data, length, 0, sa_tmp, sizeof(*sa_tmp)) == -1){
    
    perror("net->send: couldn't send");
    abort();
    exit(-6);
    
    }else{
        
        return 1;
        
    }

}

/*------------------------------------------------------------------*/

/**
 * net_socket *socket_new();
 *
 *   Creates a net_socket and returns it.
 */
 
net_socket *socket_new(net *n){
    
    return net_socket_new();
  
}

/**
 * This is used to find out if a specific address belongs to the host the
 * program is running on. In Linux if finds addresses from the /etc/hosts file
 * and from DNS. It will only work if 'multi on' is in the /etc/host.conf file.
 *
 * @param n The net object.
 * @param addr The address to check for.
 * @return 1 if the address belongs to us, 0 otherwize.
 */
static int is_self(net *n, unsigned long addr) {
    struct hostent *he;
    char buf[100];
    int len, i = 0;
    
    len = 100;
    gethostname(buf, len);
    he = gethostbyname(buf);
    if(he != NULL) {
        while((struct in_addr *)he->h_addr_list[i]){
            if(((struct in_addr *)he->h_addr_list[i])->s_addr == addr){
                return 1;
            }
            i++;
        }
        return 0;
    } else {
        fprintf(stderr, "Could not do reverse lookup of hostname %s\n", buf);
        return 1;
    }
}

/*------------------------------------------------------------------*/

/**
 * net *net_new(tim *timer);
 *
 *   Constructor function for the net object.
 *   This function return a pointer to a new,
 *   created net object. The argument of the
 *   net_new function is the timer object to
 *   use .
 */
net *net_new(tim *timer){
    
    net *n;
    n = (net *)malloc(sizeof(net));
    n->sender = (net_peer *)malloc(sizeof(net_peer));
    n->registrations = list_new();
    
    /* initialize the net object */
    n->receive = receive;
    n->unreceive = unreceive;
    n->send = send_msg;
    n->main = main_loop;
    n->timer = timer; 
    n->is_self = is_self;
    n->socket = socket_new;
    return n;
}
