#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "net_socket.h"
#include "debug.h"

/**
 * net_socket.c
 *
 * Implements a network connection abstraction. See net_socket.h for
 * details.
 */

static int set_destination(net_socket *ns, char *ip_addr, int port);
static int delete_ns(net_socket *ns);
static int bind_socket(net_socket *ns, char *ip_addr, int port);
static int pub_join_multicast_group(net_socket *ns,
                                    unsigned long ip_address, unsigned int ttl);

/*------------------------------------------------------------------*/
/**
 * int set_destination(net_socket *ns, char *ip_addr, int port);
 *
 *   Implements the set_dest method. See net_socket.h for details.
 */

static int set_destination(net_socket *ns, char *ip_addr, int port){

  bzero((char *) &ns->sa_dest, sizeof(ns->sa_dest)); /* just in case */
  
  ns->sa_dest.sin_family = AF_INET;
  
  ns->sa_dest.sin_addr.s_addr = inet_addr(ip_addr);
  ns->sa_dest.sin_port = htons( port ); /* Connect to port  */

 return 1;

}

/*------------------------------------------------------------------*/
/**
 * int delete_ns(net_socket *ns);
 *
 *   Implements the delete method. See net_socket.h for details.
 *
 */
static int delete_ns(net_socket *ns){

  close(ns->socket);
  return 1;
}

/*------------------------------------------------------------------*/
/**
 * int bind_socket(net_socket *ns, char *ip_addr, int port);
 *
 *   Implements the bind method. See net_socket.h for details.
 *
 */
static int bind_socket(net_socket *ns, char *ip_addr, int port){

  struct sockaddr_in sa;
  
  bzero( (char *)&sa , sizeof(sa) );
  
  sa.sin_family = AF_INET;

  /* check if we should bind to any interface */
  if( ip_addr == NET_SOCKET_ANY ){
    /* bind to any interface */
    sa.sin_addr.s_addr = htonl(INADDR_ANY);
  }else{
    struct hostent *he;
    he = gethostbyname(ip_addr);
    if(he != NULL) {
      sa.sin_family = he->h_addrtype;
      bcopy(he->h_addr, (char *)&sa.sin_addr, he->h_length);
    } else {
      printf("Host %s not found.\n",ip_addr);
      exit(-1);
    }


    /* check if we should bind to a multicast address,
       and if so, join the multicast group */
    if( ntohl(sa.sin_addr.s_addr) >= ntohl(inet_addr("224.0.0.0")) ){
      struct ip_mreq ipmq;
      unsigned char loop;
      /* specify which multicast group to join,
	 the same as given as argument to the bind method */
      ipmq.imr_multiaddr.s_addr = sa.sin_addr.s_addr;
      /* we should join with any (kernel chosen) interface */
      ipmq.imr_interface.s_addr = INADDR_ANY;

      /* join the group defined in ipmq */
      setsockopt(ns->socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, &ipmq,
                 sizeof(ipmq));

      loop = 0;
      setsockopt(ns->socket, IPPROTO_IP, IP_MULTICAST_LOOP, &loop,
                 sizeof(loop));

    }
	
  }
  
  sa.sin_port = htons( port );

  /* finally, bind the actual socket (ns->socket) */
  if( bind( ns->socket , (struct sockaddr *) &sa , sizeof(sa)) < 0){
    fprintf(stderr, "Can't bind socket to %s:%d", ip_addr, port);
    perror(" ");
    exit(-2);
  }  
  return 1;
}

/**
 * Function called to join multicastgroups.
 * @param ns The netsocket that should be used to receive multicast datagrams.
 * @param ip_address The multicast address(in host byte order)
 * that should be joined. 
 * @param ttl The number of routers to traverse.
 * @return -1 on failure, otherwise 0.
 */
static int pub_join_multicast_group(net_socket *ns,
                                    unsigned long ip_address, unsigned int ttl){

    struct ip_mreq ipmq;
    unsigned char loop  = 0;
    int ret_val = 0;
    char multicast_group[50];
    struct sockaddr_in tmp;
    int pktinfo_optval = 1;
    
    bzero((char *)&ipmq , sizeof(ipmq));
    bzero((char *)&tmp , sizeof(tmp));
    
    /* Join the multicast group. */    
    ipmq.imr_multiaddr.s_addr = htonl(ip_address);
    ipmq.imr_interface.s_addr = INADDR_ANY;
    ret_val = setsockopt(ns->socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, &ipmq,
                         sizeof(ipmq));
    if (ret_val == -1){
        return -1;
    }

    /* Set TTL to traverse up to multiple routers. */
    ret_val = setsockopt(ns->socket, IPPROTO_IP, IP_MULTICAST_TTL, &ttl,
                         sizeof(ttl));
    if (ret_val == -1){
       return -1;
    }

    /* Disable loopback. */
    loop = 0;
    ret_val = setsockopt(ns->socket, IPPROTO_IP, IP_MULTICAST_LOOP, &loop,
                         sizeof(loop));
    if (ret_val == -1){
       return -1;
    }

    if(setsockopt(ns->socket, SOL_IP, IP_PKTINFO, &pktinfo_optval,
                  sizeof(int)) == -1) {
        return -1;
    }

    tmp.sin_addr.s_addr = ipmq.imr_multiaddr.s_addr; 
    sprintf(multicast_group,"Joined multicast group %s with TTL %d.%s",
            inet_ntoa(tmp.sin_addr), ttl, "\n");    
    DEBUG('n',multicast_group);
    
    return 0;
}

/*------------------------------------------------------------------*/
static unsigned long name(net_socket *ns) {
  struct sockaddr_in sa;
  int len;
  bzero((char *)&sa , sizeof(sa));

  len = sizeof(sa);
  if(getsockname(ns->socket, (struct sockaddr *)&sa, &len) != -1) {
    return sa.sin_addr.s_addr;
  }
  perror("getsockname:");
  return -1;
}

/*------------------------------------------------------------------*/
/**
 * net_socket *net_socket_new();
 *
 *   Creates a new net_socket object.
 *
 */

net_socket *net_socket_new() {  

  int sckt;
  
  net_socket *ns = (net_socket *)malloc(sizeof(net_socket));

  /* create a UDP socket */
  if( (sckt = socket(AF_INET,SOCK_DGRAM,0)) < 0){
    perror("Can't create socket");
    exit(-1);
  }

  /*setsockopt(sckt,SOL_SOCKET,SO_REUSEADDR,NULL,NULL);*/
  
  ns->set_dest = set_destination;
  ns->socket = sckt;
  ns->delete = delete_ns;
  ns->bind = bind_socket;
  ns->name = name;
  ns->join_multicast_group = pub_join_multicast_group;
  return ns;

}
